import itertools
import json
import math
import random
import time
from collections import defaultdict
from collections.abc import Sized
from itertools import chain, repeat
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
)

import spacy
import spacy.tokenizer
import torch
from accelerate import Accelerator
from confit import Cli
from confit.registry import validate_arguments
from confit.utils.random import set_seed
from pydantic import BaseModel
from rich_logger import RichTablePrinter
from spacy.tokens import Doc
from tqdm import tqdm

import edsnlp
from edsnlp.core.pipeline import Pipeline
from edsnlp.core.registries import registry
from edsnlp.optimization import LinearSchedule, ScheduledOptimizer
from edsnlp.pipes.trainable.embeddings.transformer.transformer import Transformer
from edsnlp.scorers import Scorer
from edsnlp.utils.bindings import BINDING_SETTERS
from edsnlp.utils.collections import batchify
from edsnlp.utils.span_getters import get_spans
from edsnlp.utils.typing import AsList

app = Cli(pretty_exceptions_show_locals=False)

LOGGER_FIELDS = {
    "step": {},
    "(.*)loss": {
        "goal": "lower_is_better",
        "format": "{:.2e}",
        "goal_wait": 2,
    },
    "exact_ner/micro/(f|r|p)$": {
        "goal": "higher_is_better",
        "format": "{:.2%}",
        "goal_wait": 1,
        "name": r"ner_\1",
    },
    "qualifier/micro/(f|r|p)$": {
        "goal": "higher_is_better",
        "format": "{:.2%}",
        "goal_wait": 1,
        "name": r"qual_\1",
    },
    "lr": {"format": "{:.2e}"},
    "speed/(.*)": {"format": "{:.2f}", r"name": r"\1"},
    "labels": {"format": "{:.2f}"},
}


def flatten_dict(d, path=""):
    if not isinstance(d, dict):
        return {path: d}

    return {
        k: v
        for key, val in d.items()
        for k, v in flatten_dict(val, f"{path}/{key}" if path else key).items()
    }


class BatchSizeArg:
    """
    Batch size argument validator / caster for confit/pydantic

    Examples
    --------

    ```{ .python .no-check }
    def fn(batch_size: BatchSizeArg):
        return batch_size


    print(fn("10 samples"))
    # Out: (10, "samples")

    print(fn("10 words"))
    # Out: (10, "words")

    print(fn(10))
    # Out: (10, "samples")
    ```
    """

    @classmethod
    def validate(cls, value, config=None):
        value = str(value)
        parts = value.split()
        num = int(parts[0])
        unit = parts[1] if len(parts) == 2 else "samples"
        if (
            len(parts) == 2
            and str(num) == parts[0]
            and unit in ("words", "samples", "spans")
        ):
            return num, unit
        raise Exception(
            f"Invalid batch size: {value}, must be <int> samples|words|spans"
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate


if TYPE_CHECKING:
    BatchSizeArg = Tuple[int, str]  # noqa: F811


class LengthSortedBatchSampler:
    """
    Batch sampler that sorts the dataset by length and then batches
    sequences of similar length together. This is useful for transformer
    models that can then be padded more efficiently.

    Parameters
    ----------
    dataset: Iterable
        The dataset to sample from (can be a generator or a fixed size collection)
    batch_size: int
        The batch size
    batch_unit: str
        The unit of the batch size, either "words" or "samples"
    noise: int
        The amount of noise to add to the sequence length before sorting
        (uniformly sampled in [-noise, noise])
    drop_last: bool
        Whether to drop the last batch if it is smaller than the batch size
    buffer_size: Optional[int]
        The size of the buffer to use to shuffle the batches. If None, the buffer
        will be approximately the size of the dataset.
    """

    def __init__(
        self,
        dataset,
        batch_size: int,
        batch_unit: str,
        noise=1,
        drop_last=True,
        buffer_size: Optional[int] = None,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.batch_unit = batch_unit
        self.noise = noise
        self.drop_last = drop_last
        self.buffer_size = buffer_size

    def __iter__(self):
        # Shuffle the dataset
        if self.batch_unit == "words":

            def sample_len(idx, noise=True):
                count = sum(
                    len(x)
                    for x in next(
                        v
                        for k, v in self.dataset[idx].items()
                        if k.endswith("word_lengths")
                    )
                )
                if not noise:
                    return count
                return count + random.randint(-self.noise, self.noise)

        elif self.batch_unit == "spans":

            def sample_len(idx, noise=True):
                return len(
                    next(
                        v for k, v in self.dataset[idx].items() if k.endswith("begins")
                    )
                )

        else:
            sample_len = lambda idx, noise=True: 1  # noqa: E731

        def make_batches():
            total = 0
            batch = []
            for seq_size, idx in sorted_sequences:
                if total and total + seq_size > self.batch_size:
                    yield batch
                    total = 0
                    batch = []
                total += seq_size
                batch.append(idx)

        # Shuffle the batches in buffer that contain approximately
        # the full dataset to add more randomness
        if isinstance(self.dataset, Sized):
            total_count = sum(sample_len(i, False) for i in range(len(self.dataset)))

        assert (
            isinstance(self.dataset, Sized) or self.buffer_size is not None
        ), "Dataset must have a length or buffer_size must be specified"
        buffer_size = self.buffer_size or math.ceil(total_count / self.batch_size)

        # Sort sequences by length +- some noise
        sorted_sequences = chain.from_iterable(
            sorted((sample_len(i), i) for i in range(len(self.dataset)))
            for _ in repeat(None)
        )

        # Batch sorted sequences
        batches = make_batches()
        buffers = batchify(batches, buffer_size)
        for buffer in buffers:
            random.shuffle(buffer)
            yield from buffer


class SubBatchCollater:
    """
    Collater that splits batches into sub-batches of a maximum size

    Parameters
    ----------
    nlp: Pipeline
        The pipeline object
    embedding: Transformer
        The transformer embedding pipe
    grad_accumulation_max_tokens: int
        The maximum number of tokens (word pieces) to accumulate in a single batch
    """

    def __init__(self, nlp, embedding, grad_accumulation_max_tokens):
        self.nlp = nlp
        self.embedding: Transformer = embedding
        self.grad_accumulation_max_tokens = grad_accumulation_max_tokens

    def __call__(self, seq):
        total = 0
        mini_batches = [[]]
        for sample_features in seq:
            num_tokens = sum(
                math.ceil(len(p) / self.embedding.stride) * self.embedding.window
                for key in sample_features
                if key.endswith("/input_ids")
                for p in sample_features[key]
            )
            if total + num_tokens > self.grad_accumulation_max_tokens:
                print(
                    f"Mini batch size was becoming too large: {total} > "
                    f"{self.grad_accumulation_max_tokens} so it was split"
                )
                total = 0
                mini_batches.append([])
            total += num_tokens
            mini_batches[-1].append(sample_features)
        return [self.nlp.collate(b) for b in mini_batches]


def subset_doc(doc: Doc, start: int, end: int) -> Doc:
    """
    Subset a doc given a start and end index.

    Parameters
    ----------
    doc: Doc
        The doc to subset
    start: int
        The start index
    end: int
        The end index

    Returns
    -------
    Doc
    """
    # TODO: review user_data copy strategy
    new_doc = doc[start:end].as_doc(copy_user_data=True)
    new_doc.user_data.update(doc.user_data)

    for name, group in doc.spans.items():
        new_doc.spans[name] = [
            spacy.tokens.Span(
                new_doc,
                max(0, span.start - start),
                min(end, span.end) - start,
                span.label,
            )
            for span in group
            if span.end > start and span.start < end
        ]

    return new_doc


class Reader(BaseModel, arbitrary_types_allowed=True):
    """
    Reader that reads docs from a file or a generator, and adapts them to the pipeline.

    Parameters
    ----------
    reader: Callable[..., Iterable[Doc]]
        The reader object
    limit: Optional[int]
        The maximum number of docs to read
    max_length: int
        The maximum length of the resulting docs
    randomize: bool
        Whether to randomize the split
    multi_sentence: bool
        Whether to split sentences across multiple docs
    filter_expr: Optional[str]
        An expression to filter the docs to generate
    """

    reader: Any
    limit: Optional[int] = -1
    max_length: int = 0
    randomize: bool = False
    multi_sentence: bool = True
    filter_expr: Optional[str] = None

    def __call__(self, nlp) -> List[Doc]:
        filter_fn = eval(f"lambda doc:{self.filter_expr}") if self.filter_expr else None

        blank_nlp = edsnlp.Pipeline(nlp.lang, vocab=nlp.vocab, vocab_config=None)
        blank_nlp.add_pipe("eds.normalizer")
        blank_nlp.add_pipe("eds.sentences")

        docs = blank_nlp.pipe(self.reader)

        count = 0

        # Load the jsonl data from path
        if self.randomize:
            docs: List[Doc] = list(docs)
            random.shuffle(docs)

        for doc in docs:
            if 0 <= self.limit <= count:
                break
            if not (len(doc) and (filter_fn is None or filter_fn(doc))):
                continue
            count += 1

            for sub_doc in self.split_doc(doc):
                if len(sub_doc.text.strip()):
                    yield sub_doc
            else:
                continue

    def split_doc(
        self,
        doc: Doc,
    ) -> Iterable[Doc]:
        """
        Split a doc into multiple docs of max_length tokens.

        Parameters
        ----------
        doc: Doc
            The doc to split

        Returns
        -------
        Iterable[Doc]
        """
        max_length = self.max_length
        randomize = self.randomize

        if max_length <= 0:
            yield doc
        else:
            start = 0
            end = 0
            for ent in doc.ents:
                for token in ent:
                    token.is_sent_start = False
            for sent in doc.sents if doc.has_annotation("SENT_START") else (doc[:],):
                # If the sentence adds too many tokens
                if sent.end - start > max_length:
                    # But the current buffer too large
                    while sent.end - start > max_length:
                        subset_end = start + int(
                            max_length * (random.random() ** 0.3 if randomize else 1)
                        )
                        yield subset_doc(doc, start, subset_end)
                        start = subset_end
                    yield subset_doc(doc, start, sent.end)
                    start = sent.end

                if not self.multi_sentence:
                    yield subset_doc(doc, start, sent.end)
                    start = sent.end

                # Otherwise, extend the current buffer
                end = sent.end

            yield subset_doc(doc, start, end)


@validate_arguments
class GenericScorer:
    def __init__(
        self,
        ner: Dict[str, Scorer] = {},
        qualifier: Dict[str, Scorer] = {},
    ):
        self.ner_scorers = ner
        self.qlf_scorers = qualifier

    def __call__(self, nlp: Pipeline, docs):
        scores = {}
        docs = list(docs)

        # Speed
        t0 = time.time()
        list(nlp.pipe(d.text for d in tqdm(docs, desc="Computing model speed")))
        duration = time.time() - t0
        scores["speed"] = dict(
            wps=sum(len(d) for d in docs) / duration,
            dps=len(docs) / duration,
        )

        # NER
        if nlp.has_pipe("ner"):
            clean_ner_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            for d in clean_ner_docs:
                d.ents = []
                d.spans.clear()
            with nlp.select_pipes(enable=["ner"]):
                ner_preds = list(nlp.pipe(tqdm(clean_ner_docs, desc="Predicting")))
            for name, scorer in self.ner_scorers.items():
                scores[name] = scorer(docs, ner_preds)

        # Qualification
        if nlp.has_pipe("qualifier"):
            qlf_pipe = nlp.get_pipe("qualifier")
            clean_qlf_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            for doc in clean_qlf_docs:
                for span in get_spans(doc, qlf_pipe.span_getter):
                    for qlf in qlf_pipe.qualifiers:
                        BINDING_SETTERS[(qlf, None)](span)
            with nlp.select_pipes(enable=["qualifier"]):
                qlf_preds = list(nlp.pipe(tqdm(clean_qlf_docs, desc="Predicting")))
            for name, scorer in self.qlf_scorers.items():
                scores[name] = scorer(docs, qlf_preds)
        return scores


@app.command(name="train", registry=registry)
def train(
    *,
    output_path: Path = Path("model"),
    nlp: Pipeline,
    train_data: AsList[Reader],
    val_data: AsList[Reader],
    seed: int = 42,
    data_seed: int = 42,
    max_steps: int = 1000,
    batch_size: BatchSizeArg = 2000,
    transformer_lr: float = 5e-5,
    task_lr: float = 3e-4,
    validation_interval: int = 10,
    max_grad_norm: float = 5.0,
    warmup_rate: float = 0.1,
    scorer: GenericScorer = GenericScorer(),
    grad_accumulation_max_tokens: int = 96 * 128,
    cpu: bool = False,
):
    trf_pipe = next(
        module
        for name, pipe in nlp.torch_components()
        for module_name, module in pipe.named_component_modules()
        if isinstance(module, Transformer)
    )

    set_seed(seed)
    with RichTablePrinter(LOGGER_FIELDS, auto_refresh=False) as logger:
        # Loading and adapting the training and validation data
        with set_seed(data_seed):
            train_docs: List[spacy.tokens.Doc] = list(
                chain.from_iterable(td(nlp) for td in train_data)
            )
            val_docs: List[spacy.tokens.Doc] = list(
                chain.from_iterable(td(nlp) for td in val_data)
            )

        # Initialize pipeline with training documents
        nlp.post_init(train_docs)

        # Preprocessing training data
        preprocessed = list(nlp.preprocess_many(train_docs, supervision=True))
        dataloader = torch.utils.data.DataLoader(
            preprocessed,
            batch_sampler=LengthSortedBatchSampler(
                preprocessed,
                batch_size=batch_size[0],
                batch_unit=batch_size[1],
            ),
            collate_fn=SubBatchCollater(
                nlp,
                trf_pipe,
                grad_accumulation_max_tokens=grad_accumulation_max_tokens,
            ),
        )
        pipe_names, trained_pipes = zip(*nlp.torch_components())
        print("Training", ", ".join(pipe_names))

        trf_params = set(trf_pipe.parameters())
        params = set(p for pipe in trained_pipes for p in pipe.parameters())
        optimizer = ScheduledOptimizer(
            torch.optim.AdamW(
                [
                    {
                        "params": list(params - trf_params),
                        "lr": task_lr,
                        "schedules": LinearSchedule(
                            total_steps=max_steps,
                            warmup_rate=warmup_rate,
                            start_value=task_lr,
                        ),
                    },
                    {
                        "params": list(trf_params),
                        "lr": transformer_lr,
                        "schedules": LinearSchedule(
                            total_steps=max_steps,
                            warmup_rate=warmup_rate,
                            start_value=0,
                        ),
                    },
                ][: 2 if transformer_lr else 1]
            )
        )
        grad_params = {p for group in optimizer.param_groups for p in group["params"]}
        print(
            "Optimizing:"
            + "".join(
                f"\n - {len(group['params'])} params "
                f"({sum(p.numel() for p in group['params'])} total)"
                for group in optimizer.param_groups
            )
        )
        print(f"Not optimizing {len(params - grad_params)} params")
        for param in params - grad_params:
            param.requires_grad_(False)

        accelerator = Accelerator(cpu=cpu)
        print("Device:", accelerator.device)
        [dataloader, optimizer, *trained_pipes] = accelerator.prepare(
            dataloader,
            optimizer,
            *trained_pipes,
        )

        cumulated_data = defaultdict(lambda: 0.0, count=0)

        iterator = itertools.chain.from_iterable(itertools.repeat(dataloader))
        all_metrics = []
        nlp.train(True)
        set_seed(seed)
        with tqdm(
            range(max_steps + 1),
            "Training model",
            leave=True,
            mininterval=5.0,
        ) as bar:
            for step in bar:
                if (step % validation_interval) == 0:
                    scores = scorer(nlp, val_docs)
                    all_metrics.append(
                        {
                            "step": step,
                            "lr": optimizer.param_groups[0]["lr"],
                            **cumulated_data,
                            **scores,
                        }
                    )
                    cumulated_data = defaultdict(lambda: 0.0, count=0)
                    nlp.to_disk(output_path)
                    (output_path / "train_metrics.json").write_text(
                        json.dumps(all_metrics, indent=2)
                    )
                    logger.log_metrics(flatten_dict(all_metrics[-1]))
                if step == max_steps:
                    break
                mini_batches = next(iterator)
                optimizer.zero_grad()
                for mini_batch in mini_batches:
                    loss = torch.zeros((), device=accelerator.device)
                    with nlp.cache():
                        for name, pipe in zip(pipe_names, trained_pipes):
                            output = pipe.module_forward(mini_batch[name])
                            if "loss" in output:
                                loss += output["loss"]
                            for key, value in output.items():
                                if key.endswith("loss"):
                                    cumulated_data[key] += float(value)
                            if torch.isnan(loss):
                                raise ValueError(f"NaN loss at component {name}")

                    accelerator.backward(loss)

                torch.nn.utils.clip_grad_norm_(grad_params, max_grad_norm)
                optimizer.step()

    return nlp


if __name__ == "__main__":
    app()



"""
Adapted from pymdownx.tabbed (https://github.com/facelessuser/pymdown-extensions/)
"""
import re
import xml.etree.ElementTree as etree

from markdown import Extension
from markdown.blockprocessors import BlockProcessor
from markdown.extensions.attr_list import AttrListTreeprocessor, get_attrs


def assign_attrs(elem, attrs):
    """Assign `attrs` to element."""
    for k, v in get_attrs(attrs):
        if k == ".":
            # add to class
            cls = elem.get("class")
            if cls:
                elem.set("class", "{} {}".format(cls, v))
            else:
                elem.set("class", v)
        else:
            # assign attribute `k` with `v`
            elem.set(AttrListTreeprocessor.NAME_RE.sub("_", k), v)


class CardProcessor(BlockProcessor):
    """card block processor."""

    START = re.compile(r"(?:^|\n)={3} *(card)?(?: +({:.*?}) *(?:\n|$))?")
    COMPRESS_SPACES = re.compile(r" {2,}")

    def __init__(self, parser, config):
        """Initialize."""

        super().__init__(parser)
        self.card_group_count = 0
        self.current_sibling = None
        self.content_indention = 0

    def detab_by_length(self, text, length):
        """Remove a card from the front of each line of the given text."""

        newtext = []
        lines = text.split("\n")
        for line in lines:
            if line.startswith(" " * length):
                newtext.append(line[length:])
            elif not line.strip():
                newtext.append("")  # pragma: no cover
            else:
                break
        return "\n".join(newtext), "\n".join(lines[len(newtext) :])

    def parse_content(self, parent, block):
        """
        Get sibling card.

        Retrieve the appropriate sibling element. This can get tricky when
        dealing with lists.

        """

        old_block = block
        non_cards = ""
        card_set = "card-set"

        # We already acquired the block via test
        if self.current_sibling is not None:
            sibling = self.current_sibling
            block, non_cards = self.detab_by_length(block, self.content_indent)
            self.current_sibling = None
            self.content_indent = 0
            return sibling, block, non_cards

        sibling = self.lastChild(parent)

        if (
            sibling is None
            or sibling.tag.lower() != "div"
            or sibling.attrib.get("class", "") != card_set
        ):
            sibling = None
        else:
            # If the last child is a list and the content is indented sufficient
            # to be under it, then the content's is sibling is in the list.
            last_child = self.lastChild(sibling)
            card_content = "card-content"
            child_class = (
                last_child.attrib.get("class", "") if last_child is not None else ""
            )
            indent = 0
            while last_child is not None:
                if (
                    sibling is not None
                    and block.startswith(" " * self.tab_length * 2)
                    and last_child is not None
                    and (
                        last_child.tag in ("ul", "ol", "dl")
                        or (last_child.tag == "div" and child_class == card_content)
                    )
                ):
                    # Handle nested card content
                    if last_child.tag == "div" and child_class == card_content:
                        temp_child = self.lastChild(last_child)
                        if temp_child is None or temp_child.tag not in (
                            "ul",
                            "ol",
                            "dl",
                        ):
                            break
                        last_child = temp_child

                    # The expectation is that we'll find an `<li>`.
                    # We should get it's last child as well.
                    sibling = self.lastChild(last_child)
                    last_child = (
                        self.lastChild(sibling) if sibling is not None else None
                    )
                    child_class = (
                        last_child.attrib.get("class", "")
                        if last_child is not None
                        else ""
                    )

                    # Context has been lost at this point, so we must adjust the
                    # text's indentation level so it will be evaluated correctly
                    # under the list.
                    block = block[self.tab_length :]
                    indent += self.tab_length
                else:
                    last_child = None

            if not block.startswith(" " * self.tab_length):
                sibling = None

            if sibling is not None:
                indent += self.tab_length
                block, non_cards = self.detab_by_length(old_block, indent)
                self.current_sibling = sibling
                self.content_indent = indent

        return sibling, block, non_cards

    def test(self, parent, block):
        """Test block."""

        if self.START.search(block):
            return True
        else:
            return self.parse_content(parent, block)[0] is not None

    def run(self, parent, blocks):
        """Convert to card block."""

        block = blocks.pop(0)
        m = self.START.search(block)
        card_set = "card-set"

        if m:
            # removes the first line
            if m.start() > 0:
                self.parser.parseBlocks(parent, [block[: m.start()]])
            block = block[m.end() :]
            sibling = self.lastChild(parent)
            block, non_cards = self.detab(block)
        else:
            sibling, block, non_cards = self.parse_content(parent, block)

        if m:
            if (
                sibling is not None
                and sibling.tag.lower() == "div"
                and sibling.attrib.get("class", "") == card_set
            ):
                card_group = sibling
            else:
                self.card_group_count += 1
                card_group = etree.SubElement(
                    parent,
                    "div",
                    {
                        "class": card_set,
                        "data-cards": "%d:0" % self.card_group_count,
                    },
                )

            data = card_group.attrib["data-cards"].split(":")
            card_set = int(data[0])
            card_count = int(data[1]) + 1

            div = etree.SubElement(
                card_group,
                "div",
                {
                    "class": "card-content",
                },
            )
            attributes = m.group(2)

            if attributes:
                attr_m = AttrListTreeprocessor.INLINE_RE.search(attributes)
                if attr_m:
                    assign_attrs(div, attr_m.group(1))
                    if div.get("href"):
                        div.tag = "a"

            card_group.attrib["data-cards"] = "%d:%d" % (card_set, card_count)
        else:
            if sibling.tag in ("li", "dd") and sibling.text:
                # Sibling is a list item, but we need to wrap it's content should be
                # wrapped in <p>
                text = sibling.text
                sibling.text = ""
                p = etree.SubElement(sibling, "p")
                p.text = text
                div = sibling
            elif sibling.tag == "div" and sibling.attrib.get("class", "") == card_set:
                # Get `card-content` under `card-set`
                div = self.lastChild(sibling)
            else:
                # Pass anything else as the parent
                div = sibling

        self.parser.parseChunk(div, block)

        if non_cards:
            # Insert the card content back into blocks
            blocks.insert(0, non_cards)


class CardExtension(Extension):
    """Add card extension."""

    def __init__(self, *args, **kwargs):
        """Initialize."""

        self.config = {
            "slugify": [
                0,
                "Slugify function used to create card specific IDs - Default: None",
            ],
            "combine_header_slug": [
                False,
                "Combine the card slug with the slug of the parent header - "
                "Default: False",
            ],
            "separator": ["-", "Slug separator - Default: '-'"],
        }

        super(CardExtension, self).__init__(*args, **kwargs)

    def extendMarkdown(self, md):
        """Add card to Markdown instance."""
        md.registerExtension(self)

        config = self.getConfigs()

        self.card_processor = CardProcessor(md.parser, config)

        md.parser.blockprocessors.register(
            self.card_processor,
            "card",
            105,
        )

    def reset(self):
        """Reset."""

        self.card_processor.card_group_count = 0


def makeExtension(*args, **kwargs):
    """Return extension."""

    return CardExtension(*args, **kwargs)



#!/usr/bin/env python
from distutils.sysconfig import get_python_inc

import numpy
from setuptools import Extension, setup

# See if Cython is installed
try:
    from Cython.Build import cythonize
    from Cython.Distutils import build_ext
except ImportError:
    # Do nothing if Cython is not available
    # Got to provide this function. Otherwise, poetry will fail
    print("You must install Cython to build this library")
else:
    # Cython is installed. Compile
    print("Compiling")

    COMPILER_DIRECTIVES = {
        "language_level": "3",
    }
    MOD_NAMES = [
        "edsnlp.matchers.phrase",
        "edsnlp.pipes.core.sentences.fast_sentences",
    ]

    include_dirs = [
        numpy.get_include(),
        get_python_inc(plat_specific=True),
    ]
    ext_modules = []
    for name in MOD_NAMES:
        mod_path = name.replace(".", "/") + ".pyx"
        ext = Extension(
            name,
            [mod_path],
            language="c++",
            include_dirs=include_dirs,
            extra_compile_args=["-std=c++11"],
        )
        ext_modules.append(ext)
    print("Cythonizing sources")
    ext_modules = cythonize(ext_modules, compiler_directives=COMPILER_DIRECTIVES)

    setup(
        ext_modules=ext_modules,
        package_data={
            "": ["*.dylib", "*.so"],
        },
        cmdclass={"build_ext": build_ext},
    )



import pytest
from confit import validate_arguments
from confit.errors import ConfitValidationError

from edsnlp.utils.typing import AsList


def test_as_list():
    @validate_arguments
    def func(a: AsList[int]):
        return a

    assert func("1") == [1]

    with pytest.raises(ConfitValidationError) as e:
        func("a")

    assert (
        "1 validation error for test_typing.test_as_list.<locals>.func()\n" "-> a.0\n"
    ) in str(e.value)



import sys

import pytest
from extract_docs_code import extract_docs_code

# We don't check documentation for Python <= 3.7:
if sys.version_info < (3, 8):
    url_to_code = {}
else:
    url_to_code = dict(extract_docs_code())


def printer(code: str) -> None:
    """
    Prints a code bloc with lines for easier debugging.

    Parameters
    ----------
    code : str
        Code bloc.
    """
    lines = []
    for i, line in enumerate(code.split("\n")):
        lines.append(f"{i + 1:03}  {line}")

    print("\n".join(lines))


# Note the use of `str`, makes for pretty output
@pytest.mark.parametrize("url", sorted(url_to_code.keys()), ids=str)
def test_code_blocks(url):
    raw = url_to_code[url]
    try:
        exec(raw, {"__MODULE__": "__main__"})
    except Exception:
        printer(raw)
        raise



from datetime import datetime
from typing import Any, Dict, List

import pandas as pd
import pytest
from pyspark.sql import types as T
from pyspark.sql.session import SparkSession
from spacy.tokens import Doc

from edsnlp.processing import pipe

text = """
Motif :
Le patient est admis le 29 août 2020 pour des difficultés respiratoires.

Antécédents familiaux :
Le père est asthmatique, sans traitement particulier.

HISTOIRE DE LA MALADIE
Le patient dit avoir de la toux. \
Elle a empiré jusqu'à nécessiter un passage aux urgences.
La patiente avait un SOFA à l'admission de 8.

Conclusion
Possible infection au coronavirus
"""


def note(module):
    data = [(i, i // 5, text, datetime(2021, 1, 1)) for i in range(20)]

    if module == "pandas":
        return pd.DataFrame(
            data=data, columns=["note_id", "person_id", "note_text", "note_datetime"]
        )

    note_schema = T.StructType(
        [
            T.StructField("note_id", T.IntegerType()),
            T.StructField("person_id", T.IntegerType()),
            T.StructField("note_text", T.StringType()),
            T.StructField(
                "note_datetime",
                T.TimestampType(),
            ),
        ]
    )

    spark = SparkSession.builder.getOrCreate()
    notes = spark.createDataFrame(data=data, schema=note_schema)
    if module == "pyspark":
        return notes

    if module == "koalas":
        return notes.to_koalas()


@pytest.fixture
def model(blank_nlp):
    # Creates the spaCy instance
    nlp = blank_nlp

    # Normalisation of accents, case and other special characters
    nlp.add_pipe("eds.normalizer")

    # Extraction of named entities
    nlp.add_pipe(
        "eds.matcher",
        config=dict(
            terms=dict(
                respiratoire=[
                    "difficultes respiratoires",
                    "asthmatique",
                    "toux",
                ]
            ),
            regex=dict(
                covid=r"(?i)(?:infection\sau\s)?(covid[\s\-]?19|corona[\s\-]?virus)",
                traitement=r"(?i)traitements?|medicaments?",
                respiratoire="respiratoires",
            ),
            attr="NORM",
        ),
    )

    # Qualification of matched entities
    nlp.add_pipe("eds.negation")
    nlp.add_pipe("eds.hypothesis")
    nlp.add_pipe("eds.family")
    nlp.add_pipe("eds.reported_speech")
    nlp.add_pipe("eds.sofa")
    nlp.add_pipe("eds.dates")

    return nlp


params = [
    dict(module="pandas", n_jobs=1),
    dict(module="pandas", n_jobs=-2),
    dict(module="pyspark", n_jobs=None),
]

try:
    import databricks.koalas  # noqa F401

    params.append(dict(module="koalas", n_jobs=None))
except ImportError:
    pass


@pytest.mark.parametrize("param", params)
def test_pipelines(param, model):
    module = param["module"]

    note_nlp = pipe(
        note(module=module),
        nlp=model,
        n_jobs=param["n_jobs"],
        context=["note_datetime"],
        extensions={
            "score_method": T.StringType(),
            "negation": T.BooleanType(),
            "hypothesis": T.BooleanType(),
            "family": T.BooleanType(),
            "reported_speech": T.BooleanType(),
            "date.year": T.IntegerType(),
            "date.month": T.IntegerType(),
        },
        additional_spans=["dates"],
    )

    if module == "pyspark":
        note_nlp = note_nlp.toPandas()
    elif module == "koalas":
        note_nlp = note_nlp.to_pandas()

    assert len(note_nlp) == 140
    assert set(note_nlp.columns) == set(
        (
            "note_id",
            "lexical_variant",
            "label",
            "span_type",
            "start",
            "end",
            "negation",
            "hypothesis",
            "reported_speech",
            "family",
            "score_method",
            "date_year",
            "date_month",
        )
    )


def test_spark_missing_types(model):
    with pytest.warns(Warning) as warned:
        pipe(
            note(module="pyspark"),
            nlp=model,
            extensions={"negation", "hypothesis", "family"},
        )
    assert any(
        "The following schema was inferred" in str(warning.message)
        for warning in warned
    )


@pytest.mark.parametrize("param", params)
def test_arbitrary_callback(param, model):
    # We need to test PySpark with an installed function
    def dummy_extractor(doc: Doc) -> List[Dict[str, Any]]:
        return [
            dict(
                snippet=ent.text,
                length=len(ent.text),
                note_datetime=doc._.note_datetime,
            )
            for ent in doc.ents
        ]

    module = param["module"]

    note_nlp = pipe(
        note(module=module),
        nlp=model,
        n_jobs=param["n_jobs"],
        context=["note_datetime"],
        results_extractor=dummy_extractor,
        dtypes={
            "snippet": T.StringType(),
            "length": T.IntegerType(),
        },
    )

    if module == "pandas":
        assert set(note_nlp.columns) == {"snippet", "length", "note_datetime"}
        assert (note_nlp.snippet.str.len() == note_nlp.length).all()

    else:
        if module == "pyspark":
            note_nlp = note_nlp.toPandas()
        elif module == "koalas":
            note_nlp = note_nlp.to_pandas()

        assert set(note_nlp.columns) == {
            "note_id",
            "snippet",
            "length",
        }
        assert (note_nlp.snippet.str.len() == note_nlp.length).all()



from edsnlp.viz.quick_examples import QuickExample


def test_quick_example_text(nlp, capsys):
    text = "Le patient présente une anomalie"

    E = QuickExample(nlp, extensions=["sent"])
    E(text)

    captured = capsys.readouterr()

    assert "Entity" in captured.out
    assert "Source" in captured.out
    assert "patient" in captured.out
    assert "anomalie" in captured.out
    assert "sent" in captured.out


def test_quick_example_doc(nlp, capsys):
    text = "Le patient présente une anomalie"

    E = QuickExample(nlp, extensions=["sent"])
    E(nlp(text))

    captured = capsys.readouterr()

    assert "Entity" in captured.out
    assert "Source" in captured.out
    assert "patient" in captured.out
    assert "anomalie" in captured.out
    assert "sent" in captured.out


def test_quick_example_df(nlp):
    text = "Le patient présente une anomalie"

    E = QuickExample(nlp, extensions=["sent"])
    df = E(text, as_dataframe=True)
    assert df.shape == (2, 8)



from datetime import datetime

import spacy
from pytest import mark

from edsnlp.pipelines.qualifiers.history import History
from edsnlp.pipelines.qualifiers.history.patterns import history
from edsnlp.pipelines.terminations import termination

text = """COMPTE RENDU D'HOSPITALISATION du 11/07/2018 au 12/07/2018

MOTIF D'HOSPITALISATION
Monsieur Dupont 30\n2 Jean Michel, de sexe masculin, âgée de 39 ans,
née le 23/11/1978, est admis pour une toux.
Il a été hospitalisé du 11/08/2019 au 17/08/2019,
avec un antécédent d'asthme il y a 25 jours.

ANTÉCÉDENTS
Antécédents médicaux :
Premier épisode: il a été hospitalisé pour asthme cette semaine-ci,
il y a 3 jours, le 13 août 2020.
Hier, le patient est venu pour une toux dont les symptômes,
seraient apparus il y a 2 mois."""


@mark.parametrize("use_sections", [True, False])
@mark.parametrize("use_dates", [True, False])
@mark.parametrize("on_ents_only", [True, False])
@mark.parametrize("exclude_birthdate", [True, False])
def test_history(lang, use_sections, use_dates, exclude_birthdate, on_ents_only):
    nlp = spacy.blank(lang)
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe(
        "eds.matcher",
        config=dict(
            terms=dict(
                respiratoire=[
                    "asthmatique",
                    "asthme",
                    "toux",
                ]
            )
        ),
    )
    nlp.add_pipe(
        "eds.history", config=dict(use_sections=use_sections, use_dates=use_dates)
    )
    doc = nlp(text)
    nlp.remove_pipe("eds.history")
    nlp.add_pipe("eds.sections")
    nlp.add_pipe("eds.dates")
    doc = nlp(text)
    doc._.note_datetime = datetime(2020, 8, 11)
    doc._.birth_datetime = datetime(1978, 11, 23)

    history_nlp = History(
        nlp=nlp,
        attr="NORM",
        history=history,
        termination=termination,
        use_sections=use_sections,
        use_dates=use_dates,
        exclude_birthdate=exclude_birthdate,
        closest_dates_only=True,
        history_limit=15,
        explain=True,
        on_ents_only=on_ents_only,
    )
    doc = history_nlp(doc)

    if use_dates:
        assert doc.ents[0]._.history is not exclude_birthdate
        if not exclude_birthdate:
            assert doc.ents[0]._.history_cues[0].text == "23/11/1978"

    assert doc.ents[1]._.history and doc.ents[1]._.history_cues[0].label_ == "history"

    if use_sections:
        assert doc.ents[2]._.history is not use_dates
        assert doc.ents[2]._.history_cues[0].label_ == "ATCD"
        if use_dates:
            assert doc.ents[2]._.recent_cues[0].label_ == "relative_date"



from typing import TYPE_CHECKING

from edsnlp.utils.lazy_module import lazify

lazify()

if TYPE_CHECKING:
    from .core.contextual_matcher.factory import create_component as contextual_matcher
    from .core.endlines.factory import create_component as endlines
    from .core.matcher.factory import create_component as matcher
    from .core.normalizer.accents.factory import create_component as accents
    from .core.normalizer.factory import create_component as normalizer
    from .core.normalizer.pollution.factory import create_component as pollution
    from .core.normalizer.quotes.factory import create_component as quotes
    from .core.normalizer.remove_lowercase.factory import (
        create_component as remove_lowercase,
    )
    from .core.normalizer.spaces.factory import create_component as spaces
    from .core.sentences.factory import create_component as sentences
    from .core.terminology.factory import create_component as terminology
    from .misc.consultation_dates.factory import create_component as consultation_dates
    from .misc.dates.factory import create_component as dates
    from .misc.measurements.factory import create_component as measurements
    from .misc.reason.factory import create_component as reason
    from .misc.sections.factory import create_component as sections
    from .misc.tables.factory import create_component as tables
    from .ner.adicap.factory import create_component as adicap
    from .ner.behaviors.alcohol.factory import create_component as alcohol
    from .ner.behaviors.tobacco.factory import create_component as tobacco
    from .ner.cim10.factory import create_component as cim10
    from .ner.covid.factory import create_component as covid
    from .ner.disorders.aids.factory import create_component as aids
    from .ner.disorders.cerebrovascular_accident.factory import (
        create_component as cerebrovascular_accident,
    )
    from .ner.disorders.ckd.factory import create_component as ckd
    from .ner.disorders.congestive_heart_failure.factory import (
        create_component as congestive_heart_failure,
    )
    from .ner.disorders.connective_tissue_disease.factory import (
        create_component as connective_tissue_disease,
    )
    from .ner.disorders.copd.factory import create_component as copd
    from .ner.disorders.dementia.factory import create_component as dementia
    from .ner.disorders.diabetes.factory import create_component as diabetes
    from .ner.disorders.hemiplegia.factory import create_component as hemiplegia
    from .ner.disorders.leukemia.factory import create_component as leukemia
    from .ner.disorders.liver_disease.factory import create_component as liver_disease
    from .ner.disorders.lymphoma.factory import create_component as lymphoma
    from .ner.disorders.myocardial_infarction.factory import (
        create_component as myocardial_infarction,
    )
    from .ner.disorders.peptic_ulcer_disease.factory import (
        create_component as peptic_ulcer_disease,
    )
    from .ner.disorders.peripheral_vascular_disease.factory import (
        create_component as peripheral_vascular_disease,
    )
    from .ner.disorders.solid_tumor.factory import create_component as solid_tumor
    from .ner.drugs.factory import create_component as drugs
    from .ner.scores.charlson.factory import create_component as charlson
    from .ner.scores.elston_ellis.factory import create_component as elston_ellis
    from .ner.scores.emergency.ccmu.factory import create_component as emergency_ccmu
    from .ner.scores.emergency.gemsa.factory import create_component as emergency_gemsa
    from .ner.scores.emergency.priority.factory import create_component as emergency_priority
    from .ner.scores.factory import create_component as score
    from .ner.scores.sofa.factory import create_component as sofa
    from .ner.tnm.factory import create_component as tnm
    from .ner.umls.factory import create_component as umls
    from .qualifiers.family.factory import create_component as family
    from .qualifiers.history.factory import create_component as history
    from .qualifiers.hypothesis.factory import create_component as hypothesis
    from .qualifiers.negation.factory import create_component as negation
    from .qualifiers.reported_speech.factory import create_component as reported_speech
    from .qualifiers.reported_speech.factory import create_component as rspeech
    from .trainable.ner_crf.factory import create_component as ner_crf
    from .trainable.span_qualifier.factory import create_component as span_qualifier
    from .trainable.span_linker.factory import create_component as span_linker
    from .trainable.embeddings.span_pooler.factory import create_component as span_pooler
    from .trainable.embeddings.transformer.factory import create_component as transformer
    from .trainable.embeddings.text_cnn.factory import create_component as text_cnn



results_peripheral_vascular_disease = dict(
    has_match=[
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Un AOMI",
        "Présence d'un infarctus rénal",
        "Une angiopathie cérébrale",
        "Une angiopathie",
        "Une thrombose cérébrale",
        "Une thrombose des veines superficielles",
        "Une thrombose",
        "Effectuer un bilan pre-trombose",
        "Une ischémie des MI est remarquée.",
        "Plusieurs cas d'EP",
        "Effectuer des cures d'EP",
        "Le patient est hypertendu",  # Echange plasmatique
        "Une hypertension portale",
    ],
)



from edsnlp.core import registry

from .reason import ReasonMatcher

DEFAULT_CONFIG = dict(
    reasons=None,
    attr="TEXT",
    use_sections=False,
    ignore_excluded=False,
)

create_component = registry.factory.register(
    "eds.reason",
    assigns=["doc.spans", "doc.ents"],
    deprecated=["reason"],
)(ReasonMatcher)



import json
from pathlib import Path

import pytest

import edsnlp


def assert_doc_read(doc):
    assert doc._.note_id == "subfolder/doc-1"
    assert doc._.context_var == "test"

    attrs = ("etat", "assertion")
    spans_and_attributes = {
        "__ents__": sorted(
            [
                (e.start, e.end, e.text, tuple(getattr(e._, key) for key in attrs))
                for e in doc.ents
            ]
        ),
        **{
            name: sorted(
                [
                    (e.start, e.end, e.text, tuple(getattr(e._, key) for key in attrs))
                    for e in doc.spans[name]
                ]
            )
            for name in doc.spans
        },
    }

    assert spans_and_attributes == {
        "__ents__": [
            (6, 7, "douleurs", (None, None)),
            (7, 11, "dans le bras droit", (None, None)),
            (17, 21, "problème \nde locomotion", (None, "absent")),
            (25, 26, "AVC", ("passé", "non-associé")),
            (35, 36, "rhume", ("présent", "hypothétique")),
            (45, 46, "rhume", ("présent", "hypothétique")),
            (51, 52, "Douleurs", (None, None)),
            (52, 56, "dans le bras droit", (None, None)),
            (68, 69, "anomalie", (None, "absent")),
        ],
        "anatomie": [
            (9, 11, "bras droit", (None, None)),
            (54, 56, "bras droit", (None, None)),
        ],
        "localisation": [
            (7, 11, "dans le bras droit", (None, None)),
            (52, 56, "dans le bras droit", (None, None)),
        ],
        "pathologie": [
            (17, 21, "problème \nde locomotion", (None, "absent")),
            (25, 26, "AVC", ("passé", "non-associé")),
            (35, 36, "rhume", ("présent", "hypothétique")),
            (45, 46, "rhume", ("présent", "hypothétique")),
        ],
        "sosy": [
            (6, 7, "douleurs", (None, None)),
            (51, 52, "Douleurs", (None, None)),
            (68, 69, "anomalie", (None, "absent")),
        ],
    }


def assert_doc_write(exported_obj):
    assert exported_obj == {
        "entities": [
            {
                "assertion": None,
                "end_char": 38,
                "etat": "test",
                "lexical_variant": "douleurs",
                "note_nlp_id": 0,
                "note_nlp_source_value": "sosy",
                "start_char": 30,
            },
            {
                "assertion": None,
                "end_char": 57,
                "etat": None,
                "lexical_variant": "dans le bras droit",
                "note_nlp_id": 1,
                "note_nlp_source_value": "localisation",
                "start_char": 39,
            },
            {
                "assertion": None,
                "end_char": 57,
                "etat": None,
                "lexical_variant": "bras droit",
                "note_nlp_id": 2,
                "note_nlp_source_value": "anatomie",
                "start_char": 47,
            },
            {
                "assertion": "absent",
                "end_char": 98,
                "etat": None,
                "lexical_variant": "problème \nde locomotion",
                "note_nlp_id": 3,
                "note_nlp_source_value": "pathologie",
                "start_char": 75,
            },
            {
                "assertion": "non-associé",
                "end_char": 117,
                "etat": "passé",
                "lexical_variant": "AVC",
                "note_nlp_id": 4,
                "note_nlp_source_value": "pathologie",
                "start_char": 114,
            },
            {
                "assertion": "hypothétique",
                "end_char": 164,
                "etat": "présent",
                "lexical_variant": "rhume",
                "note_nlp_id": 5,
                "note_nlp_source_value": "pathologie",
                "start_char": 159,
            },
            {
                "assertion": "hypothétique",
                "end_char": 296,
                "etat": "présent",
                "lexical_variant": "rhume",
                "note_nlp_id": 6,
                "note_nlp_source_value": "pathologie",
                "start_char": 291,
            },
            {
                "assertion": None,
                "end_char": 314,
                "etat": None,
                "lexical_variant": "Douleurs",
                "note_nlp_id": 7,
                "note_nlp_source_value": "sosy",
                "start_char": 306,
            },
            {
                "assertion": None,
                "end_char": 333,
                "etat": None,
                "lexical_variant": "dans le bras droit",
                "note_nlp_id": 8,
                "note_nlp_source_value": "localisation",
                "start_char": 315,
            },
            {
                "assertion": None,
                "end_char": 333,
                "etat": None,
                "lexical_variant": "bras droit",
                "note_nlp_id": 9,
                "note_nlp_source_value": "anatomie",
                "start_char": 323,
            },
            {
                "assertion": "absent",
                "end_char": 386,
                "etat": None,
                "lexical_variant": "anomalie",
                "note_nlp_id": 10,
                "note_nlp_source_value": "sosy",
                "start_char": 378,
            },
        ],
        "note_id": "subfolder/doc-1",
        "context_var": "test",
        "note_text": "Le patient est admis pour des douleurs dans le bras droit, mais "
        "n'a pas de problème \n"
        "de locomotion. \n"
        "Historique d'AVC dans la famille. pourrait être un cas de "
        "rhume.\n"
        "NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNBWbWbNbNbNBWbNbWbNbWBNbNbWbNbNBNbWb"
        "WbNbWBNbNbWbNBNbWbWbNb\n"
        "Pourrait être un cas de rhume.\n"
        "Motif :\n"
        "Douleurs dans le bras droit.\n"
        "ANTÉCÉDENTS\n"
        "Le patient est déjà venu\n"
        "Pas d'anomalie détectée.\n",
    }


def test_read_in_worker(blank_nlp, tmpdir):
    input_dir = Path(__file__).parent.parent.resolve() / "resources" / "docs.jsonl"
    list(
        edsnlp.data.read_json(
            input_dir,
            converter="omop",
            span_attributes=["etat", "assertion"],
            read_in_worker=True,
        )
    )[0]


def test_read_to_json(blank_nlp, tmpdir):
    input_dir = Path(__file__).parent.parent.resolve() / "resources" / "docs.jsonl"
    output_dir = Path(tmpdir)
    doc = list(
        edsnlp.data.read_json(
            input_dir,
            converter="omop",
            span_attributes=["etat", "assertion"],
            doc_attributes=["context_var"],
        )
    )[0]
    assert_doc_read(doc)
    doc.ents[0]._.etat = "test"

    edsnlp.data.write_json(
        [doc],
        output_dir / "docs.jsonl",
        converter="omop",
        doc_attributes=["context_var"],
        span_attributes=["etat", "assertion"],
        span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        lines=True,
    )

    with pytest.raises(FileExistsError):
        edsnlp.data.write_json(
            [doc],
            output_dir / "docs.jsonl",
            converter="omop",
            doc_attributes=["context_var"],
            span_attributes=["etat", "assertion"],
            span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        )

    with open(output_dir / "docs.jsonl") as f:
        exported_obj = json.loads(f.readlines()[0])
    assert_doc_write(exported_obj)

    with pytest.raises(FileExistsError):
        edsnlp.data.write_json(
            [doc],
            output_dir,
            converter="omop",
            doc_attributes=["context_var"],
            span_attributes=["etat", "assertion"],
            span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
            lines=False,
        )

    edsnlp.data.write_json(
        [doc],
        output_dir,
        converter="omop",
        doc_attributes=["context_var"],
        span_attributes=["etat", "assertion"],
        span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        lines=False,
        overwrite=True,
    )

    with open(output_dir / "subfolder" / "doc-1.json") as f:
        exported_obj = json.loads(f.read())
    assert_doc_write(exported_obj)



from pathlib import Path

import pyarrow.dataset
import pyarrow.fs
import pytest

import edsnlp
from edsnlp.data.converters import get_dict2doc_converter, get_doc2dict_converter
from edsnlp.utils.collections import dl_to_ld


def assert_doc_read(doc):
    assert doc._.note_id == "subfolder/doc-1"
    assert doc._.context_var == "test"

    attrs = ("etat", "assertion")
    spans_and_attributes = {
        "__ents__": sorted(
            [
                (e.start, e.end, e.text, tuple(getattr(e._, key) for key in attrs))
                for e in doc.ents
            ]
        ),
        **{
            name: sorted(
                [
                    (e.start, e.end, e.text, tuple(getattr(e._, key) for key in attrs))
                    for e in doc.spans[name]
                ]
            )
            for name in doc.spans
        },
    }

    assert spans_and_attributes == {
        "__ents__": [
            (6, 7, "douleurs", (None, None)),
            (7, 11, "dans le bras droit", (None, None)),
            (17, 21, "problème \nde locomotion", (None, "absent")),
            (25, 26, "AVC", ("passé", "non-associé")),
            (35, 36, "rhume", ("présent", "hypothétique")),
            (45, 46, "rhume", ("présent", "hypothétique")),
            (51, 52, "Douleurs", (None, None)),
            (52, 56, "dans le bras droit", (None, None)),
            (68, 69, "anomalie", (None, "absent")),
        ],
        "anatomie": [
            (9, 11, "bras droit", (None, None)),
            (54, 56, "bras droit", (None, None)),
        ],
        "localisation": [
            (7, 11, "dans le bras droit", (None, None)),
            (52, 56, "dans le bras droit", (None, None)),
        ],
        "pathologie": [
            (17, 21, "problème \nde locomotion", (None, "absent")),
            (25, 26, "AVC", ("passé", "non-associé")),
            (35, 36, "rhume", ("présent", "hypothétique")),
            (45, 46, "rhume", ("présent", "hypothétique")),
        ],
        "sosy": [
            (6, 7, "douleurs", (None, None)),
            (51, 52, "Douleurs", (None, None)),
            (68, 69, "anomalie", (None, "absent")),
        ],
    }


GOLD_OMOP = {
    "entities": [
        {
            "assertion": None,
            "end_char": 38,
            "etat": "test",
            "lexical_variant": "douleurs",
            "note_nlp_id": 0,
            "note_nlp_source_value": "sosy",
            "start_char": 30,
        },
        {
            "assertion": None,
            "end_char": 57,
            "etat": None,
            "lexical_variant": "dans le bras droit",
            "note_nlp_id": 1,
            "note_nlp_source_value": "localisation",
            "start_char": 39,
        },
        {
            "assertion": None,
            "end_char": 57,
            "etat": None,
            "lexical_variant": "bras droit",
            "note_nlp_id": 2,
            "note_nlp_source_value": "anatomie",
            "start_char": 47,
        },
        {
            "assertion": "absent",
            "end_char": 98,
            "etat": None,
            "lexical_variant": "problème \nde locomotion",
            "note_nlp_id": 3,
            "note_nlp_source_value": "pathologie",
            "start_char": 75,
        },
        {
            "assertion": "non-associé",
            "end_char": 117,
            "etat": "passé",
            "lexical_variant": "AVC",
            "note_nlp_id": 4,
            "note_nlp_source_value": "pathologie",
            "start_char": 114,
        },
        {
            "assertion": "hypothétique",
            "end_char": 164,
            "etat": "présent",
            "lexical_variant": "rhume",
            "note_nlp_id": 5,
            "note_nlp_source_value": "pathologie",
            "start_char": 159,
        },
        {
            "assertion": "hypothétique",
            "end_char": 296,
            "etat": "présent",
            "lexical_variant": "rhume",
            "note_nlp_id": 6,
            "note_nlp_source_value": "pathologie",
            "start_char": 291,
        },
        {
            "assertion": None,
            "end_char": 314,
            "etat": None,
            "lexical_variant": "Douleurs",
            "note_nlp_id": 7,
            "note_nlp_source_value": "sosy",
            "start_char": 306,
        },
        {
            "assertion": None,
            "end_char": 333,
            "etat": None,
            "lexical_variant": "dans le bras droit",
            "note_nlp_id": 8,
            "note_nlp_source_value": "localisation",
            "start_char": 315,
        },
        {
            "assertion": None,
            "end_char": 333,
            "etat": None,
            "lexical_variant": "bras droit",
            "note_nlp_id": 9,
            "note_nlp_source_value": "anatomie",
            "start_char": 323,
        },
        {
            "assertion": "absent",
            "end_char": 386,
            "etat": None,
            "lexical_variant": "anomalie",
            "note_nlp_id": 10,
            "note_nlp_source_value": "sosy",
            "start_char": 378,
        },
    ],
    "note_id": "subfolder/doc-1",
    "context_var": "test",
    "note_text": "Le patient est admis pour des douleurs dans le bras droit, mais "
    "n'a pas de problème \n"
    "de locomotion. \n"
    "Historique d'AVC dans la famille. pourrait être un cas de "
    "rhume.\n"
    "NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNBWbWbNbNbNBWbNbWbNbWBNbNbWbNbNBNbWb"
    "WbNbWBNbNbWbNBNbWbWbNb\n"
    "Pourrait être un cas de rhume.\n"
    "Motif :\n"
    "Douleurs dans le bras droit.\n"
    "ANTÉCÉDENTS\n"
    "Le patient est déjà venu\n"
    "Pas d'anomalie détectée.\n",
}


def assert_doc_write_omop(exported_obj):
    assert exported_obj == GOLD_OMOP


def assert_doc_write_ents(exported_objs):
    in_converter, kwargs = get_dict2doc_converter(
        "omop",
        dict(
            span_attributes=["etat", "assertion"],
            doc_attributes=["context_var"],
        ),
    )
    doc = in_converter(GOLD_OMOP, **kwargs)
    out_converter, kwargs = get_doc2dict_converter(
        "ents",
        dict(
            span_attributes=["etat", "assertion"],
            doc_attributes=["context_var"],
            span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        ),
    )
    GOLD_ENTS = out_converter(doc, **kwargs)
    assert exported_objs == GOLD_ENTS


def test_read_write_in_worker(blank_nlp, tmpdir):
    input_dir = Path(__file__).parent.parent.resolve() / "resources" / "docs.pq"
    output_dir = Path(tmpdir)
    edsnlp.data.read_parquet(
        input_dir,
        converter="omop",
        span_attributes=["etat", "assertion"],
        doc_attributes=["context_var"],
        read_in_worker=True,
    ).write_parquet(
        output_dir / "docs.pq",
        converter="omop",
        doc_attributes=["context_var"],
        span_attributes=["etat", "assertion"],
        span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        write_in_worker=True,
    )
    # fmt: off
    assert (
          list(dl_to_ld(pyarrow.dataset.dataset(output_dir / "docs.pq").to_table().to_pydict()))  # noqa: E501
          == list(dl_to_ld(pyarrow.dataset.dataset(input_dir).to_table().to_pydict()))
    )
    # fmt: on


def test_read_to_parquet(blank_nlp, tmpdir):
    input_dir = Path(__file__).parent.parent.resolve() / "resources" / "docs.pq"
    output_dir = Path(tmpdir)
    fs = pyarrow.fs.LocalFileSystem()
    doc = list(
        edsnlp.data.read_parquet(
            input_dir,
            converter="omop",
            span_attributes=["etat", "assertion"],
            doc_attributes=["context_var"],
            filesystem=fs,
        )
    )[0]
    assert_doc_read(doc)
    doc.ents[0]._.etat = "test"

    edsnlp.data.write_parquet(
        [doc],
        output_dir,
        converter="omop",
        doc_attributes=["context_var"],
        span_attributes=["etat", "assertion"],
        span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
    )

    assert_doc_write_omop(
        next(dl_to_ld(pyarrow.dataset.dataset(output_dir).to_table().to_pydict()))
    )

    with pytest.raises(FileExistsError):
        edsnlp.data.write_parquet(
            [doc],
            output_dir,
            converter="omop",
            doc_attributes=["context_var"],
            span_attributes=["etat", "assertion"],
            span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        )

    edsnlp.data.write_parquet(
        [doc],
        output_dir,
        converter="omop",
        doc_attributes=["context_var"],
        span_attributes=["etat", "assertion"],
        span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        overwrite=True,
    )


def test_read_to_parquet_ents(blank_nlp, tmpdir):
    input_dir = Path(__file__).parent.parent.resolve() / "resources" / "docs.pq"
    output_dir = Path(tmpdir)
    fs = pyarrow.fs.LocalFileSystem()
    doc = list(
        edsnlp.data.read_parquet(
            input_dir,
            converter="omop",
            span_attributes=["etat", "assertion"],
            doc_attributes=["context_var"],
            filesystem=fs,
        )
    )[0]
    assert_doc_read(doc)
    doc.ents[0]._.etat = "test"

    edsnlp.data.write_parquet(
        [doc],
        output_dir,
        converter="ents",
        doc_attributes=["context_var"],
        span_attributes=["etat", "assertion"],
        span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
    )

    assert_doc_write_ents(
        list(dl_to_ld(pyarrow.dataset.dataset(output_dir).to_table().to_pydict()))
    )

    with pytest.raises(FileExistsError):
        edsnlp.data.write_parquet(
            [doc],
            output_dir,
            converter="ents",
            doc_attributes=["context_var"],
            span_attributes=["etat", "assertion"],
            span_getter=["ents", "sosy", "localisation", "anatomie", "pathologie"],
        )



from spacy.tokens import Span


def check_inclusion(span: Span, start: int, end: int) -> bool:
    """
    Checks whether the span overlaps the boundaries.

    Parameters
    ----------
    span : Span
        Span to check.
    start : int
        Start of the boundary
    end : int
        End of the boundary

    Returns
    -------
    bool
        Whether the span overlaps the boundaries.
    """

    if span.start >= end or span.end <= start:
        return False
    return True


def check_sent_inclusion(span: Span, start: int, end: int) -> bool:
    """
    Checks whether the span overlaps the boundaries.

    Parameters
    ----------
    span : Span
        Span to check.
    start : int
        Start of the boundary
    end : int
        End of the boundary

    Returns
    -------
    bool
        Whether the span overlaps the boundaries.
    """
    if span.sent.start >= end or span.sent.end <= start:
        return False
    return True



results_aids = dict(
    has_match=[
        True,
        False,
        True,
        True,
    ],
    detailled_status=[
        "PRESENT",
        None,
        "PRESENT",
        "PRESENT",
    ],
    assign=None,
    texts=[
        "Patient atteint du VIH au stade SIDA.",
        "Patient atteint du VIH.",
        "Il y a un VIH avec coinfection pneumocystose",
        "Présence d'un VIH stade C",
    ],
)



from edsnlp.core import registry

from .patterns import default_enabled
from .pollution import PollutionTagger

DEFAULT_CONFIG = dict(
    pollution=default_enabled,
)

create_component = registry.factory.register(
    "eds.pollution",
    assigns=["doc.spans"],
    deprecated=["pollution"],
)(PollutionTagger)



import os
import warnings
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Tuple

import pystow
from tqdm import tqdm
from umls_downloader import download_umls

PATTERN_VERSION = "0.7.0"
UMLS_VERSION = "2022AA"


def get_patterns(config: Dict[str, Any]) -> Dict[str, List[str]]:
    """Load the UMLS terminology patterns.

    Parameters
    ----------
    config : dict[list]
        Languages and sources to select from the whole terminology.
        For both keys, None will select all values.

    Return
    ------
    patterns : dict[list]
        The mapping between CUI codes and their synonyms.

    Notes
    -----
    When run for the first time, this method will download the
    entire UMLS file and store it at ~/.data/bio/umls/2022AA/.
    Therefore the second run will be significantly faster than
    the first one.
    """

    path, module, filename = get_path(config)

    if path.exists():  # pragma: no cover
        print(f"Loading {filename} from {module.base}")
        return module.load_pickle(name=filename)
    else:  # pragma: no cover
        patterns = download_and_agg_umls(config)
        module.dump_pickle(name=filename, obj=patterns)
        print(f"Saved patterns into {module.base / filename}")
        return patterns


def get_path(config: Dict[str, Any]) -> Tuple[Path, pystow.impl.Module, str]:
    """Get the path, module and filename of the UMLS file.

    Parameters
    ----------
    config : dict[list]
        Languages and sources to select from the whole terminology.
        For both keys, None will select all values.

    Return
    ------
    path, module, filename : pathlib.Path, pystow.module, str

    Notes
    -----
    `get_path` will convert the config dict into a pretty filename.

    Examples
    --------
    >>> config = {"languages": ["FRE", "ENG"], "sources": None}
    >>> print(get_path(config))
    .data/bio/umls/2022AA/languagesFRE-ENG_sourcesNone.pkl"
    """
    config_txt = ""
    for k, v in config.items():
        if isinstance(v, (list, tuple)):
            v = "-".join(v)
        _config_txt = f"{k}{v}"
        if config_txt:
            _config_txt = f"_{_config_txt}"
        config_txt += _config_txt

    filename = f"{PATTERN_VERSION}_{config_txt}.pkl"
    module = pystow.module("bio", "umls", UMLS_VERSION)
    path = module.base / filename

    return path, module, filename


def download_and_agg_umls(config) -> Dict[str, List[str]]:
    """Download the UMLS if not exist and create a mapping
    between CUI code and synonyms.

    Parameters
    ----------
    config : dict[list]
        Languages and sources to select from the whole terminology.
        For both keys, None will select all values.

    Return
    ------
    patterns : dict[list]
        The mapping between CUI codes and their synonyms.

    Notes
    -----
    Performs filtering on the returned mapping only, not the downloaded
    resource.
    """

    api_key = os.getenv("UMLS_API_KEY")
    if not api_key:  # pragma: no cover
        warnings.warn(
            "You need to define UMLS_API_KEY to download the UMLS. "
            "Get a key by creating an account at "
            "https://uts.nlm.nih.gov/uts/signup-login"
        )

    path = download_umls(version=UMLS_VERSION, api_key=api_key)

    # https://www.ncbi.nlm.nih.gov/books/NBK9685/table/ch03.T.concept_names_and_sources_file_mr/ # noqa
    header = [
        "CUI",
        "LAT",
        "TS",
        "LUI",
        "STT",
        "SUI",
        "ISPREF",
        "AUI",
        "SAUI",
        "SCUI",
        "SDUI",
        "SAB",
        "TTY",
        "CODE",
        "STR",
        "STRL",
        "SUPPRESS",
    ]
    header_to_idx = dict(zip(header, range(len(header))))
    patterns = defaultdict(list)

    languages = config.get("languages")
    sources = config.get("sources")

    with zipfile.ZipFile(path) as zip_file:
        with zip_file.open("MRCONSO.RRF", mode="r") as file:
            for row in tqdm(
                file.readlines(), desc="Loading 'MRCONSO.RRF' into a dictionnary"
            ):
                row = row.decode("utf-8").split("|")

                if (languages is None or row[header_to_idx["LAT"]] in languages) and (
                    sources is None or row[header_to_idx["SAB"]] in sources
                ):
                    cui = row[header_to_idx["CUI"]]
                    synonym = row[header_to_idx["STR"]]
                    patterns[cui].append(synonym)

    return patterns



import pytest
import torch

import edsnlp
from edsnlp.utils.collections import ld_to_dl


def test_map_batches():
    items = [1, 2, 3, 4, 5]
    lazy = edsnlp.data.from_iterable(items)
    lazy = lazy.map(lambda x: x + 1)
    lazy = lazy.map_batches(lambda x: [sum(x)])
    lazy = lazy.set_processing(
        num_cpu_workers=2,
        sort_chunks=False,
        batch_size=2,
    )
    res = list(lazy)
    assert set(res) == {5, 9, 6}


@pytest.mark.parametrize("num_cpu_workers", [1, 2])
def test_flat_iterable(num_cpu_workers):
    items = [1, 2, 3, 4]
    lazy = edsnlp.data.from_iterable(items)
    lazy = lazy.set_processing(num_cpu_workers=num_cpu_workers)
    lazy = lazy.map(lambda x: [x] * x)
    res = list(lazy)
    assert sorted(res) == [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]


@pytest.mark.parametrize("num_gpu_workers", [0, 1])
def test_map_gpu(num_gpu_workers):
    def prepare_batch(batch, device):
        return {"tensor": torch.tensor(batch).to(device)}

    def forward(batch):
        return {"outputs": batch["tensor"] * 2}

    items = [1, 2, 3, 4, 5]
    lazy = edsnlp.data.from_iterable(items)
    lazy = lazy.map(lambda x: x + 1)
    lazy = lazy.map_gpu(prepare_batch, forward)
    lazy = lazy.set_processing(
        num_gpu_workers=num_gpu_workers,
        gpu_worker_devices=["cpu"] * num_gpu_workers,
        sort_chunks=False,
        batch_size=2,
    )

    res = ld_to_dl(lazy)
    res = torch.cat(res["outputs"])
    assert set(res.tolist()) == {4, 6, 8, 10, 12}


@pytest.mark.parametrize("num_cpu_workers", [1, 2])
@pytest.mark.parametrize(
    "batch_by,expected",
    [
        ("words", [3, 1, 3, 1, 3, 1]),
        ("padded_words", [2, 1, 1, 2, 1, 1, 2, 1, 1]),
        ("docs", [10, 2]),
        ("ents", [3, 2, 3, 3, 1]),
    ],
)
def test_map_with_batching(num_cpu_workers, batch_by, expected):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.matcher",
        config={
            "terms": {
                "foo": ["This", "is", "a", "sentence", ".", "Short", "snippet", "too"],
            }
        },
        name="matcher",
    )
    samples = [
        "This is a sentence.",
        "Short snippet",
        "Short snippet too",
        "This is a very very long sentence that will make more than 10 words",
    ] * 3
    lazy = edsnlp.data.from_iterable(samples)
    lazy = lazy.map_pipeline(nlp)
    lazy = lazy.map_batches(len)
    lazy = lazy.to("cpu")
    lazy = lazy.set_processing(
        num_cpu_workers=num_cpu_workers,
        batch_size=10,
        batch_by=batch_by,
        chunk_size=1000,
        split_into_batches_after="matcher",
        show_progress=True,
    )
    assert list(lazy) == expected



BENINE = r"benign|benin|(grade.?\b[i1]\b)"
STAGE = r"stade ([^\s]*)"

main_pattern = dict(
    source="main",
    regex=[
        r"carcinom(?!.{0,10}in.?situ)",
        r"seminome",
        r"(?<!lympho)(?<!lympho-)sarcome",
        r"blastome",
        r"cancer([^o]|\s|\b)",
        r"adamantinome",
        r"chordome",
        r"craniopharyngiome",
        r"melanome",
        r"neoplasie",
        r"neoplasme",
        r"linite",
        r"melanome",
        r"mesoteliome",
        r"mesotheliome",
        r"seminome",
        r"myxome",
        r"paragangliome",
        r"craniopharyngiome",
        r"k .{0,5}(prostate|sein)",
        r"pancoast.?tobias",
        r"syndrome.{1,10}lynch",
        r"li.?fraumeni",
        r"germinome",
        r"adeno[\s-]?k",
        r"thymome",
        r"\bnut\b",
        r"\bgist\b",
        r"\bchc\b",
        r"\badk\b",
        r"\btves\b",
        r"\btv.tves\b",
        r"lesion.{1,20}tumor",
        r"tumeur",
        r"carcinoid",
        r"histiocytome",
        r"ependymome",
        # r"primitif", Trop de FP
    ],
    exclude=dict(
        regex=BENINE,
        window=(0, 5),
    ),
    regex_attr="NORM",
    assign=[
        dict(
            name="metastasis",
            regex=r"(metasta|multinodul)",
            window=(-3, 7),
            reduce_mode="keep_last",
        ),
        dict(
            name="stage",
            regex=STAGE,
            window=7,
            reduce_mode="keep_last",
        ),
    ],
)

metastasis_pattern = dict(
    source="metastasis",
    regex=[
        r"cellule.{1,5}tumorale.{1,5}circulantes",
        r"metasta",
        r"multinodul",
        r"carcinose",
        r"ruptures.{1,5}corticale",
        r"envahissement.{0,15}parties\smolle",
        r"(localisation|lesion)s?.{0,20}second",
        r"(lymphangite|meningite).{1,5}carcinomateuse",
    ],
    regex_attr="NORM",
    exclude=dict(
        regex=r"goitre",
        window=-3,
    ),
)

default_patterns = [
    main_pattern,
    metastasis_pattern,
]



from pytest import fixture

from edsnlp.pipelines.misc.sections import Sections, patterns
from edsnlp.utils.examples import parse_example

sections_text = (
    "Le patient est admis pour des douleurs dans le bras droit, "
    "mais n'a pas de problème de locomotion. "
    "Historique d'AVC dans la famille. pourrait être un cas de rhume.\n"
    "NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNBWbWbNbNbNBWbNb"
    "WbNbWBNbNbWbNbNBNbWbWbNbWBNbNbWbNBNbWbWbNb\n"
    "Pourrait être un cas de rhume.\n"
    "Motif :\n"
    "<ent section=motif>Douleurs</ent> dans le bras droit.\n"
    "Pas d'anomalie détectée."
)

empty_sections_text = """
Antécédents :
Conclusion :
<ent section=conclusion>Patient</ent> va mieux

Au total:
sortie du patient
"""


def test_section_detection(doc):
    assert doc.spans["sections"]


@fixture
def sections_factory(blank_nlp):

    default_config = dict(
        sections=patterns.sections,
        add_patterns=True,
        attr="NORM",
        ignore_excluded=True,
    )

    def factory(**kwargs):

        config = dict(**default_config)
        config.update(kwargs)

        return Sections(
            nlp=blank_nlp,
            **config,
        )

    return factory


def test_sections(blank_nlp, sections_factory):

    blank_nlp.add_pipe("normalizer")

    sections = sections_factory()

    text, entities = parse_example(example=sections_text)

    doc = blank_nlp(text)
    doc.ents = [
        doc.char_span(ent.start_char, ent.end_char, "generic") for ent in entities
    ]

    doc = sections(doc)

    for entity, ent in zip(entities, doc.ents):

        for modifier in entity.modifiers:

            assert (
                getattr(ent._, modifier.key) == modifier.value
            ), f"{modifier.key} labels don't match."


def test_empty_sections(blank_nlp, sections_factory):

    blank_nlp.add_pipe("normalizer")

    sections = sections_factory()

    text, entities = parse_example(example=empty_sections_text)

    doc = blank_nlp(text)
    doc.ents = [
        doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
    ]

    doc = sections(doc)

    for section in doc.spans["sections"]:
        for ent in section.ents:
            ent._.section = section.label_

    for entity, ent in zip(entities, doc.ents):
        for modifier in entity.modifiers:
            assert (
                getattr(ent._, modifier.key) == modifier.value
            ), f"{modifier.key} labels don't match."



import inspect
import warnings
from operator import attrgetter
from typing import (
    List,
    Optional,
    Tuple,
)

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.core.registries import CurriedFactory
from edsnlp.utils.span_getters import (
    SpanGetter,  # noqa: F401
    SpanGetterArg,  # noqa: F401
    SpanSetter,
    SpanSetterArg,
    get_spans,  # noqa: F401
    set_spans,
    validate_span_getter,  # noqa: F401
    validate_span_setter,
)


def value_getter(span: Span):
    key = span._._get_key("value")
    if key in span.doc.user_data:
        return span.doc.user_data[key]
    return span._.get(span.label_) if span._.has(span.label_) else None


class BaseComponentMeta(type):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)

        sig = inspect.signature(cls.__init__)
        sig = sig.replace(parameters=tuple(sig.parameters.values())[1:])
        cls.__signature__ = sig

    def __call__(cls, nlp=inspect.Signature.empty, *args, **kwargs):
        # If this component is missing the nlp argument, we curry it with the
        # provided arguments and return a CurriedFactory object.
        sig = inspect.signature(cls.__init__)
        bound = sig.bind_partial(None, nlp, *args, **kwargs)
        bound.arguments.pop("self", None)
        if (
            "nlp" in sig.parameters
            and sig.parameters["nlp"].default is sig.empty
            and bound.arguments.get("nlp", sig.empty) is sig.empty
        ):
            return CurriedFactory(cls, bound.arguments)
        if nlp is inspect.Signature.empty:
            bound.arguments.pop("nlp", None)
        return super().__call__(**bound.arguments)


class BaseComponent(metaclass=BaseComponentMeta):
    """
    The `BaseComponent` adds a `set_extensions` method,
    called at the creation of the object.

    It helps decouple the initialisation of the pipeline from
    the creation of extensions, and is particularly usefull when
    distributing EDSNLP on a cluster, since the serialisation mechanism
    imposes that the extensions be reset.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.name = name
        self.set_extensions()

    def set_extensions(self):
        """
        Set `Doc`, `Span` and `Token` extensions.
        """
        if Span.has_extension("value"):
            if Span.get_extension("value")[2] is not value_getter:
                warnings.warn(
                    "A Span extension 'value' already exists with a different getter. "
                    "Keeping the existing extension, but some components of edsnlp may "
                    "not work as expected."
                )
            return
        Span.set_extension(
            "value",
            getter=value_getter,
        )

    def get_spans(self, doc: Doc):  # noqa: F811
        """
        Returns sorted spans of interest according to the
        possible value of `on_ents_only`.
        Includes `doc.ents` by default, and adds eventual SpanGroups.
        """
        ents = list(doc.ents) + list(doc.spans.get("discarded", []))

        on_ents_only = getattr(self, "on_ents_only", None)

        if isinstance(on_ents_only, str):
            on_ents_only = [on_ents_only]
        if isinstance(on_ents_only, (set, list)):
            for spankey in set(on_ents_only) & set(doc.spans.keys()):
                ents.extend(doc.spans.get(spankey, []))

        return sorted(list(set(ents)), key=(attrgetter("start", "end")))

    def _boundaries(
        self, doc: Doc, terminations: Optional[List[Span]] = None
    ) -> List[Tuple[int, int]]:
        """
        Create sub sentences based sentences and terminations found in text.

        Parameters
        ----------
        doc:
            spaCy Doc object
        terminations:
            List of tuples with (match_id, start, end)

        Returns
        -------
        boundaries:
            List of tuples with (start, end) of spans
        """

        if terminations is None:
            terminations = []

        sent_starts = [sent.start for sent in doc.sents]
        termination_starts = [t.start for t in terminations]

        starts = sent_starts + termination_starts + [len(doc)]

        # Remove duplicates
        starts = list(set(starts))

        # Sort starts
        starts.sort()

        boundaries = [(start, end) for start, end in zip(starts[:-1], starts[1:])]

        return boundaries

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.set_extensions()


class BaseNERComponent(BaseComponent):
    span_setter: SpanSetter

    def __init__(
        self,
        nlp: PipelineProtocol = None,
        name: str = None,
        *args,
        span_setter: SpanSetterArg,
        **kwargs,
    ):
        super().__init__(nlp, name, *args, **kwargs)
        self.span_setter: SpanSetter = validate_span_setter(span_setter)  # type: ignore

    def set_spans(self, doc, matches):
        return set_spans(doc, matches, self.span_setter)



import re
from bisect import bisect_left, bisect_right
from typing import Any, Dict, List, Optional, Tuple, Union

from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.utils.doc_to_text import get_char_offsets, get_text
from edsnlp.utils.regex import compile_regex

from .utils import Patterns


def get_normalized_variant(doclike) -> str:
    tokens = [t.text + t.whitespace_ for t in doclike if not t._.excluded]
    variant = "".join(tokens)
    variant = variant.rstrip(" ")
    variant = re.sub(r"\s+", " ", variant)
    return variant


def spans_generator(match: re.Match) -> Tuple[int, int]:
    """
    Iterates over every group, and then yields the full match

    Parameters
    ----------
    match : re.Match
        A match object

    Yields
    ------
    Tuple[int, int]
        A tuple containing the start and end of the group or match
    """
    for idx in range(1, len(match.groups()) + 1):
        yield match.start(idx), match.end(idx)
    yield match.start(0), match.end(0)


def span_from_match(
    match: re.Match,
    span_from_group: bool,
) -> Tuple[int, int]:
    """
    Return the span (as a (start, end) tuple) of the first matching group.
    If `span_from_group=True`, returns the full match instead.

    Parameters
    ----------
    match : re.Match
        The Match object
    span_from_group : bool
        Whether to work on groups or on the full match

    Returns
    -------
    Tuple[int, int]
        A tuple containing the start and end of the group or match
    """
    if not span_from_group:
        start_char, end_char = match.start(), match.end()
    else:
        start_char, end_char = next(filter(lambda x: x[0] >= 0, spans_generator(match)))
    return start_char, end_char


def create_span(
    doclike: Union[Doc, Span],
    start_char: int,
    end_char: int,
    key: str,
    attr: str,
    alignment_mode: str,
    ignore_excluded: bool,
    ignore_space_tokens: bool,
) -> Optional[Span]:
    """
    spaCy only allows strict alignment mode for char_span on Spans.
    This method circumvents this.
    Parameters
    ----------
    doclike : Union[Doc, Span]
        `Doc` or `Span`.
    start_char : int
        Character index within the Doc-like object.
    end_char : int
        Character index of the end, within the Doc-like object.
    key : str
        The key used to match.
    alignment_mode : str
        The alignment mode.
    ignore_excluded : bool
        Whether to skip excluded tokens.
    ignore_space_tokens : bool
        Whether to skip space tokens.

    Returns
    -------
    span:
        A span matched on the Doc-like object.
    """

    doc = doclike if isinstance(doclike, Doc) else doclike.doc

    # Handle the simple case immediately
    if attr in {"TEXT", "LOWER"} and not ignore_excluded and not ignore_space_tokens:
        off = doclike[0].idx
        return doc.char_span(
            start_char + off,
            end_char + off,
            label=key,
            alignment_mode=alignment_mode,
        )

    clean_starts, clean_ends = get_char_offsets(
        doc,
        attr=attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
    )

    first = clean_starts[doclike[0].i]

    start_char += first
    end_char += first

    # (rightmost token that starts right after the start_char, but not at it)
    start_i = bisect_right(clean_starts, start_char)
    # (leftmost token that ends before or at the end_char) + 1
    end_i = bisect_left(clean_ends, end_char)

    if alignment_mode == "expand":
        # if the start_char is inside the previous token (<end), we want to include it
        if start_i > 0 and start_char < clean_ends[start_i - 1]:
            start_i -= 1
        # if the end_char is inside the next token (>start), we want to include it
        if end_i < len(clean_starts) and end_char > clean_starts[end_i]:
            end_i += 1
    elif alignment_mode == "strict":
        start_i -= 1
        if (
            start_i >= 0
            and start_char != clean_starts[start_i]
            or end_char != clean_ends[end_i]
        ):
            return None
        end_i += 1
    elif alignment_mode == "contract":
        # if start_char is before the previous token (<=start), we want to include it
        if start_i > 0 and start_char <= clean_starts[start_i - 1]:
            start_i -= 1
        # if end_char is after the next token (>=end), we want to include it
        if end_i < len(clean_ends) and clean_ends[end_i] <= end_char:
            end_i += 1

        # Compatibility with spacy behavior, but ideally we would return an empty span
        if end_i <= start_i:
            return None
    else:
        raise ValueError(f"Invalid alignment mode: {alignment_mode}")

    return Span(doc, start_i, max(start_i, end_i), label=key)


class RegexMatcher(object):
    """
    Simple RegExp matcher.

    Parameters
    ----------
    alignment_mode : str
        How spans should be aligned with tokens.
        Possible values are `strict` (character indices must be aligned
        with token boundaries), "contract" (span of all tokens completely
        within the character span), "expand" (span of all tokens at least
        partially covered by the character span).
        Defaults to `expand`.
    attr : str
        Default attribute to match on, by default "TEXT".
        Can be overiden in the `add` method.
    flags : Union[re.RegexFlag, int]
        Additional flags provided to the `re` module.
        Can be overiden in the `add` method.
    ignore_excluded : bool
        Whether to skip exclusions
    ignore_space_tokens: bool
        Whether to skip space tokens during matching.

        You won't be able to match on newlines if this is enabled and
        the "spaces"/"newline" option of `eds.normalizer` is enabled (by default).
    span_from_group : bool
        If set to `False`, will create spans basede on the regex's full match.
        If set to `True`, will use the first matching capturing group as a span
        (and fall back to using the full match if no capturing group is matching)
    """

    def __init__(
        self,
        alignment_mode: str = "expand",
        attr: str = "TEXT",
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        flags: Union[re.RegexFlag, int] = 0,  # No additional flags
        span_from_group: bool = False,
    ):
        self.alignment_mode = alignment_mode
        self.regex = []

        self.default_attr = attr

        self.flags = flags
        self.span_from_group = span_from_group

        self.ignore_excluded = ignore_excluded
        self.ignore_space_tokens = ignore_space_tokens

        self.set_extensions()

    @classmethod
    def set_extensions(cls):
        if not Span.has_extension("normalized_variant"):
            Span.set_extension("normalized_variant", getter=get_normalized_variant)

    def build_patterns(self, regex: Patterns):
        """
        Build patterns and adds them for matching.
        Helper function for pipelines using this matcher.

        Parameters
        ----------
        regex : Patterns
            Dictionary of label/terms, or label/dictionary of terms/attribute.
        """
        if not regex:
            regex = dict()

        for key, patterns in regex.items():
            if isinstance(patterns, dict):
                attr = patterns.get("attr")
                alignment_mode = patterns.get("alignment_mode")
                flags = patterns.get("flags")
                patterns = patterns.get("regex")
            else:
                attr = None
                alignment_mode = None
                flags = None

            if isinstance(patterns, str):
                patterns = [patterns]

            self.add(
                key=key,
                patterns=patterns,
                attr=attr,
                alignment_mode=alignment_mode,
                flags=flags,
            )

    def add(
        self,
        key: str,
        patterns: List[str],
        attr: Optional[str] = None,
        ignore_excluded: Optional[bool] = None,
        ignore_space_tokens: Optional[bool] = None,
        alignment_mode: Optional[str] = None,
        flags: Optional[re.RegexFlag] = None,
    ):
        """
        Add a pattern to the registry.

        Parameters
        ----------
        key : str
            Key of the new/updated pattern.
        patterns : List[str]
            List of patterns to add.
        attr : Optional[str]
            Attribute to use for matching.
            By default, uses the `default_attr` attribute
        ignore_excluded : Optional[bool]
            Whether to skip excluded tokens during matching.
        ignore_space_tokens: Optional[bool]
            Whether to skip space tokens during matching.

            You won't be able to match on newlines if this is enabled and
            the "spaces"/"newline" option of `eds.normalizer` is enabled (by default).

        alignment_mode : Optional[str]
            Overwrite alignment mode.
        """

        if attr is None:
            attr = self.default_attr

        if ignore_excluded is None:
            ignore_excluded = self.ignore_excluded

        if ignore_space_tokens is None:
            ignore_space_tokens = self.ignore_space_tokens

        if alignment_mode is None:
            alignment_mode = self.alignment_mode

        if flags is None:
            flags = self.flags

        patterns = [compile_regex(pattern, flags) for pattern in patterns]

        self.regex.append(
            (
                key,
                patterns,
                attr,
                ignore_excluded,
                ignore_space_tokens,
                alignment_mode,
            )
        )

    def remove(
        self,
        key: str,
    ):
        """
        Remove a pattern for the registry.

        Parameters
        ----------
        key : str
            key of the pattern to remove.

        Raises
        ------
        ValueError
            If the key is not present in the registered patterns.
        """
        n = len(self.regex)
        self.regex = [pat for pat in self.regex if pat[0] != key]
        if len(self.regex) == n:
            raise ValueError(f"`{key}` is not referenced in the matcher")

    def __len__(self):
        return len(set([regex[0] for regex in self.regex]))

    def match(
        self,
        doclike: Union[Doc, Span],
    ) -> Tuple[Span, re.Match]:
        """
        Iterates on the matches.

        Parameters
        ----------
        doclike:
            spaCy Doc or Span object to match on.

        Yields
        -------
        span:
            A match.
        """

        for (
            key,
            patterns,
            attr,
            ignore_excluded,
            ignore_space_tokens,
            alignment_mode,
        ) in self.regex:
            text = get_text(doclike, attr, ignore_excluded, ignore_space_tokens)

            for pattern in patterns:
                for match in pattern.finditer(text):
                    logger.trace(f"Matched a regex from {key}: {repr(match.group())}")

                    start_char, end_char = span_from_match(
                        match=match,
                        span_from_group=self.span_from_group,
                    )

                    span = create_span(
                        doclike=doclike,
                        start_char=start_char,
                        end_char=end_char,
                        key=key,
                        attr=attr,
                        alignment_mode=alignment_mode,
                        ignore_excluded=ignore_excluded,
                        ignore_space_tokens=ignore_space_tokens,
                    )

                    if span is None:
                        continue

                    yield span, match

    def match_with_groupdict_as_spans(
        self,
        doclike: Union[Doc, Span],
    ) -> Tuple[Span, Dict[str, Span]]:
        """
        Iterates on the matches.

        Parameters
        ----------
        doclike:
            spaCy Doc or Span object to match on.

        Yields
        -------
        span:
            A match.
        """

        for (
            key,
            patterns,
            attr,
            ignore_excluded,
            ignore_space_tokens,
            alignment_mode,
        ) in self.regex:
            text = get_text(doclike, attr, ignore_excluded, ignore_space_tokens)

            for pattern in patterns:
                for match in pattern.finditer(text):
                    logger.trace(f"Matched a regex from {key}: {repr(match.group())}")

                    start_char, end_char = span_from_match(
                        match=match,
                        span_from_group=self.span_from_group,
                    )

                    span = create_span(
                        doclike=doclike,
                        start_char=start_char,
                        end_char=end_char,
                        key=key,
                        attr=attr,
                        alignment_mode=alignment_mode,
                        ignore_excluded=ignore_excluded,
                        ignore_space_tokens=ignore_space_tokens,
                    )
                    group_spans = {}
                    for group_key, group_string in match.groupdict().items():
                        if group_string:
                            group_spans[group_key] = create_span(
                                doclike=doclike,
                                start_char=match.start(group_key),
                                end_char=match.end(group_key),
                                key=group_key,
                                attr=attr,
                                alignment_mode=alignment_mode,
                                ignore_excluded=ignore_excluded,
                                ignore_space_tokens=ignore_space_tokens,
                            )

                    yield span, group_spans

    def __call__(
        self,
        doclike: Union[Doc, Span],
        as_spans=False,
        return_groupdict=False,
    ) -> Union[Span, Tuple[Span, Dict[str, Any]]]:
        """
        Performs matching. Yields matches.

        Parameters
        ----------
        doclike:
            spaCy Doc or Span object.
        as_spans:
            Returns matches as spans.

        Yields
        ------
        span:
            A match.
        groupdict:
            Additional information coming from the named patterns
            in the regular expression.
        """
        for span, match in self.match(doclike):
            if not as_spans:
                offset = doclike[0].i
                span = (span.label, span.start - offset, span.end - offset)
            if return_groupdict:
                yield span, match.groupdict()
            else:
                yield span



from collections import defaultdict
from typing import Any, Dict, Iterable

from spacy.training import Example

from edsnlp import registry
from edsnlp.scorers import make_examples, prf
from edsnlp.utils.span_getters import SpanGetter, SpanGetterArg, get_spans


def ner_exact_scorer(
    examples: Iterable[Example],
    span_getter: SpanGetter,
    micro_key: str = "micro",
) -> Dict[str, Any]:
    """
    Scores the extracted entities that may be overlapping or nested
    by looking in the spans returned by a given SpanGetter object.

    Parameters
    ----------
    examples: Iterable[Example]
        The examples to score
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types

    Returns
    -------
    Dict[str, Any]
    """
    labels = defaultdict(lambda: (set(), set()))
    labels["micro"] = (set(), set())
    for eg_idx, eg in enumerate(examples):
        for span in (
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        ):
            labels[span.label_][0].add((eg_idx, span.start, span.end, span.label_))
            labels[micro_key][0].add((eg_idx, span.start, span.end, span.label_))

        for span in (
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        ):
            labels[span.label_][1].add((eg_idx, span.start, span.end, span.label_))
            labels[micro_key][1].add((eg_idx, span.start, span.end, span.label_))

    return {name: prf(pred, gold) for name, (pred, gold) in labels.items()}


def ner_token_scorer(
    examples: Iterable[Example],
    span_getter: SpanGetter,
    micro_key: str = "micro",
) -> Dict[str, Any]:
    """
    Scores the extracted entities that may be overlapping or nested
    by looking in `doc.ents`, and `doc.spans`, and comparing the predicted
    and gold entities at the TOKEN level.

    Parameters
    ----------
    examples: Iterable[Example]
        The examples to score
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types

    Returns
    -------
    Dict[str, Any]
    """
    # label -> pred, gold
    labels = defaultdict(lambda: (set(), set()))
    labels["micro"] = (set(), set())
    for eg_idx, eg in enumerate(examples):
        for span in (
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        ):
            for i in range(span.start, span.end):
                labels[span.label_][0].add((eg_idx, i, span.label_))
                labels[micro_key][0].add((eg_idx, i, span.label_))

        for span in (
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        ):
            for i in range(span.start, span.end):
                labels[span.label_][1].add((eg_idx, i, span.label_))
                labels[micro_key][1].add((eg_idx, i, span.label_))

    return {name: prf(pred, gold) for name, (pred, gold) in labels.items()}


@registry.scorers.register("eds.ner_exact_scorer")
def create_ner_exact_scorer(
    span_getter: SpanGetterArg,
):
    return lambda *args, **kwargs: ner_exact_scorer(
        make_examples(*args, **kwargs), span_getter
    )


@registry.scorers.register("eds.ner_token_scorer")
def create_ner_token_scorer(
    span_getter: SpanGetterArg,
):
    return lambda *args: ner_token_scorer(make_examples(*args), span_getter)



from spacy.tokens import Doc, Span

from edsnlp.utils.filter import filter_spans


def test_filter_spans(doc: Doc):
    spans = [
        doc[0:3],
        doc[0:4],
        doc[1:2],
        doc[0:2],
        doc[0:3],
    ]

    filtered = filter_spans(spans)

    assert len(filtered) == 1
    assert len(filtered[0]) == 4


def test_filter_spans_strict_nesting(doc: Doc):
    spans = [
        doc[0:5],
        doc[1:4],
    ]

    filtered = filter_spans(spans)

    assert len(filtered) == 1
    assert len(filtered[0]) == 5


def test_label_to_remove(doc: Doc):

    spans = [
        Span(doc, 0, 5, label="test"),
        Span(doc, 6, 10, label="test"),
        Span(doc, 6, 10, label="remove"),
    ]

    filtered = filter_spans(spans, label_to_remove="remove")

    assert len(filtered) == 2

    spans = [
        Span(doc, 6, 10, label="remove"),
        Span(doc, 0, 5, label="test"),
        Span(doc, 6, 10, label="test"),
    ]

    filtered = filter_spans(spans, label_to_remove="remove")

    assert len(filtered) == 1



import spacy

import edsnlp


def make_nlp(lang):
    if lang == "eds":
        model = spacy.blank("eds")
    else:
        model = edsnlp.blank("fr")

    model.add_pipe("eds.normalizer")

    model.add_pipe("eds.sentences")
    model.add_pipe("eds.sections")

    model.add_pipe(
        "eds.matcher",
        config=dict(
            terms=dict(patient="patient"),
            attr="NORM",
            ignore_excluded=True,
        ),
    )
    model.add_pipe(
        "eds.matcher",
        name="matcher2",
        config=dict(
            regex=dict(anomalie=r"anomalie"),
        ),
    )

    model.add_pipe("eds.hypothesis")
    model.add_pipe("eds.negation")
    model.add_pipe("eds.family")
    model.add_pipe("eds.history")
    model.add_pipe("eds.reported_speech")

    model.add_pipe("eds.dates")
    model.add_pipe("eds.measurements")

    return model



import gzip
import json
from functools import lru_cache
from typing import List, Optional

import pandas as pd

from edsnlp import BASE_DIR


def get_verbs(
    verbs: Optional[List[str]] = None, check_contains: bool = True
) -> pd.DataFrame:
    """
    Extract verbs from the resources, as a pandas dataframe.

    Parameters
    ----------
    verbs : List[str], optional
        List of verbs to keep. Returns all verbs by default.
    check_contains : bool, optional
        Whether to check that no verb is missing if a list of verbs was provided.
        By default True

    Returns
    -------
    pd.DataFrame
        DataFrame containing conjugated verbs.
    """

    conjugated_verbs = pd.read_csv(BASE_DIR / "resources" / "verbs.csv.gz")

    if not verbs:
        return conjugated_verbs

    verbs = set(verbs)

    selected_verbs = conjugated_verbs[conjugated_verbs.verb.isin(verbs)]

    if check_contains:
        assert len(verbs) == selected_verbs.verb.nunique(), "Some verbs are missing !"

    return selected_verbs


@lru_cache()
def get_adicap_dict():
    """
    Returns
    -------
    Dict
    """

    with gzip.open(BASE_DIR / "resources" / "adicap.json.gz", "r") as fin:
        decode_dict = json.loads(fin.read().decode("utf-8"))

    return decode_dict


@lru_cache()
def get_AVC_care_site(prefix: bool = True):
    df = pd.read_csv(BASE_DIR / "resources" / "AVC.csv.gz")

    col = (
        df.care_site_name.str.split(" ").str[:2].str.join(" ")
        if prefix
        else df.care_site_name
    )
    return set(col.to_list())



results_alcohol = dict(
    has_match=[
        True,
        True,
        False,
        False,
        True,
        True,
        True,
        True,
    ],
    detailled_status=[
        "PRESENT",
        "PRESENT",
        None,
        None,
        "ABSTINENCE",
        "PRESENT",
        "ABSENT",
        "ABSTINENCE",
    ],
    assign=None,
    texts=[
        "Patient alcoolique.",
        "OH chronique.",
        "Prise d'alcool occasionnelle",
        "Application d'un pansement alcoolisé",
        "Alcoolisme sevré",
        "Alcoolisme non sevré",
        "Alcool: 0",
        "Le patient est en cours de sevrage éthylotabagique",
    ],
)



aids = dict(
    source="aids",
    regex=[
        r"(vih.{1,5}stade.{1,5})?\bsida\b",
    ],
    regex_attr="NORM",
)

hiv = dict(
    source="hiv",
    regex=[
        r"\bhiv\b",
        r"\bvih\b",
    ],
    exclude=dict(
        regex=["serologie", "prelevement"],
        window=(-20, 20),
        limit_to_sentence=False,
    ),
    assign=[
        dict(
            name="opportunist",
            regex=r"("
            + r"|".join(
                [
                    r"kapo[sz]i",
                    r"toxoplasmose",
                    r"meningo.?encephalite.toxo",
                    r"pneumocystose",
                    r"\bpep\b",
                    r"pneumocystis",
                    r"cryptococcose",
                    r"cytomégalovirus",
                    r"myobact",
                    r"opportunist",
                    r"co.?infect",
                ]
            )
            + ")"
            + r"(?!.{0,20}(?:non|0))",
            window=(-10, 30),
            limit_to_sentence=False,
        ),
        dict(
            name="stage",
            regex=r"stade.{0,5}\b(b|c)\b",
            window=10,
        ),
    ],
    regex_attr="NORM",
)

default_patterns = [
    aids,
    hiv,
]



from edsnlp import registry

from .ner_crf import TrainableNerCrf

create_component = registry.factory.register(
    "eds.ner_crf",
    assigns=["doc.ents", "doc.spans"],
    deprecated=[
        "eds.nested_ner",
        "nested_ner",
    ],
)(TrainableNerCrf)



import filecmp
import re
from os import listdir
from os.path import join
from pathlib import Path
from random import choice, randint, random
from string import ascii_letters, ascii_lowercase

import pytest

import edsnlp
from edsnlp.connectors.brat import BratConnector
from edsnlp.core import PipelineProtocol


def random_word():
    n = randint(1, 20)
    return "".join(
        [choice(ascii_letters)] + [choice(ascii_lowercase) for _ in range(n)]
    )


def random_text():
    n = randint(30, 60)
    return " ".join([random_word() for _ in range(n)])


def random_brat_file(text):
    brat = []

    for match in re.finditer(r"\w+", text):
        if random() > 0.8:
            line = (
                f"T{len(brat) + 1}\tTEST {match.start()} {match.end()}\t{match.group()}"
            )
            brat.append(line)

    return "\n".join(brat) + "\n"


@pytest.fixture
def brat_folder(tmpdir):
    for i in range(100):
        text = random_text()
        brat = random_brat_file(text)

        with open(join(tmpdir, f"{i}.txt"), "w") as f:
            f.write(text)

        with open(join(tmpdir, f"{i}.ann"), "w") as f:
            if i == 0:
                f.write("\n")
            else:
                f.write(brat)

    return tmpdir


# OLD BratConnector tests, deprecated


@pytest.fixture
def brat1(brat_folder) -> BratConnector:
    return BratConnector(brat_folder)


@pytest.fixture
def brat2(tmpdir) -> BratConnector:
    return BratConnector(tmpdir)


@pytest.fixture
def brat_importer():
    brat_dir = Path(__file__).parent.parent.resolve() / "resources" / "brat_data"
    return BratConnector(str(brat_dir), bool_attributes=["bool flag 0"])


@pytest.fixture
def brat_exporter(tmpdir):
    return BratConnector(tmpdir, attributes=["etat", "assertion", "bool flag 0"])


def test_empty_brat(brat2: BratConnector, blank_nlp: PipelineProtocol):
    with pytest.raises(AssertionError):
        brat2.brat2docs(blank_nlp)


def test_brat2brat(
    brat1: BratConnector, brat2: BratConnector, blank_nlp: PipelineProtocol
):
    docs = brat1.brat2docs(blank_nlp)
    brat2.docs2brat(docs)

    files = listdir(brat1.directory)

    assert files

    for file in files:
        assert file in listdir(brat2.directory)
        assert filecmp.cmp(join(brat1.directory, file), join(brat2.directory, file))


def test_docs2brat(nlp, brat2):
    text = (
        "Le patient est admis pour des douleurs dans le bras droit, "
        "mais n'a pas de problème de locomotion. "
        "Historique d'AVC dans la famille. pourrait être un cas de rhume.\n"
        "NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNBWbWbNbNbN"
        "BWbNbWbNbWBNbNbWbNbNBNbWbWbNbWBNbNbWbNBNbWbWbNb\n"
        "Pourrait être un cas de rhume.\n"
        "Motif :\n"
        "Douleurs dans le bras droit."
    )

    doc1 = nlp(text)
    doc1.ents = doc1.spans["pollutions"]

    doc2 = nlp(text)
    doc2.ents = doc2.spans["section_titles"]

    docs = [doc1, doc2]
    for i, doc in enumerate(docs):
        doc._.note_id = i + 1

    brat2.docs2brat(docs)


def assert_doc_read(doc):
    assert doc._.note_id == "subfolder/doc-1"

    attrs = ("etat", "assertion", "bool flag 0")
    spans_and_attributes = {
        "__ents__": sorted(
            [
                (e.start, e.end, e.text, tuple(getattr(e._, key) for key in attrs))
                for e in doc.ents
            ]
        ),
        **{
            name: sorted(
                [
                    (e.start, e.end, e.text, tuple(getattr(e._, key) for key in attrs))
                    for e in doc.spans[name]
                ]
            )
            for name in doc.spans
        },
    }

    assert spans_and_attributes == {
        "__ents__": [
            (6, 7, "douleurs", (None, None, False)),
            (7, 11, "dans le bras droit", (None, None, False)),
            (17, 21, "problème \nde locomotion", (None, "absent", True)),
            (25, 26, "AVC", ("passé", "non-associé", False)),
            (35, 36, "rhume", ("présent", "hypothétique", False)),
            (45, 46, "rhume", ("présent", "hypothétique", False)),
            (51, 52, "Douleurs", (None, None, False)),
            (52, 56, "dans le bras droit", (None, None, False)),
            (68, 69, "anomalie", (None, "absent", False)),
        ],
        "anatomie": [
            (9, 11, "bras droit", (None, None, False)),
            (54, 56, "bras droit", (None, None, False)),
        ],
        "localisation": [
            (7, 11, "dans le bras droit", (None, None, False)),
            (52, 56, "dans le bras droit", (None, None, False)),
        ],
        "pathologie": [
            (17, 21, "problème \nde locomotion", (None, "absent", True)),
            (25, 26, "AVC", ("passé", "non-associé", False)),
            (35, 36, "rhume", ("présent", "hypothétique", False)),
            (45, 46, "rhume", ("présent", "hypothétique", False)),
        ],
        "sosy": [
            (6, 7, "douleurs", (None, None, False)),
            (51, 52, "Douleurs", (None, None, False)),
            (68, 69, "anomalie", (None, "absent", False)),
        ],
        "test label 0": [(68, 69, "anomalie", (None, "absent", False))],
    }


def assert_doc_write(exported_ann_text):
    assert exported_ann_text == (
        "T1	sosy 30 38	douleurs\n"
        "A1	etat T1 test\n"
        "T2	localisation 39 57	dans le bras droit\n"
        "T3	anatomie 47 57	bras droit\n"
        "T4	pathologie 75 83;85 98	problème de locomotion\n"
        "A2	assertion T4 absent\n"
        "A3	bool flag 0 T4\n"
        "T5	pathologie 114 117	AVC\n"
        "A4	etat T5 passé\n"
        "A5	assertion T5 non-associé\n"
        "T6	pathologie 159 164	rhume\n"
        "A6	etat T6 présent\n"
        "A7	assertion T6 hypothétique\n"
        "T7	pathologie 291 296	rhume\n"
        "A8	etat T7 présent\n"
        "A9	assertion T7 hypothétique\n"
        "T8	sosy 306 314	Douleurs\n"
        "T9	localisation 315 333	dans le bras droit\n"
        "T10	anatomie 323 333	bras droit\n"
        "T11	sosy 378 386	anomalie\n"
        "A10	assertion T11 absent\n"
        "T12	test label 0 378 386	anomalie\n"
        "A11	assertion T12 absent\n"
    )


def test_brat(
    brat_importer: BratConnector,
    brat_exporter: BratConnector,
    blank_nlp: PipelineProtocol,
):
    doc = brat_importer.brat2docs(blank_nlp)[0]
    assert_doc_read(doc)
    doc.ents[0]._.etat = "test"

    brat_exporter.docs2brat([doc])
    with open(brat_exporter.directory / "subfolder" / "doc-1.ann") as f:
        exported_ann_text = f.read()

    assert_doc_write(exported_ann_text)


# New `edsnlp.data.read_standoff` and `edsnlp.data.write_standoff` tests


def test_read_to_standoff(blank_nlp, tmpdir):
    input_dir = Path(__file__).parent.parent.resolve() / "resources" / "brat_data"
    output_dir = Path(tmpdir)
    doc = list(
        edsnlp.data.read_standoff(
            input_dir,
            bool_attributes=["bool flag 0"],
            notes_as_span_attribute="cui",
        )
    )[0]
    assert_doc_read(doc)
    doc.ents[0]._.etat = "test"
    doc.ents[0]._.cui = "C0030193"

    edsnlp.data.write_standoff(
        [doc],
        output_dir,
        span_attributes=["etat", "assertion", "bool flag 0"],
        span_getter=[
            "ents",
            "sosy",
            "localisation",
            "anatomie",
            "pathologie",
            "test label 0",
        ],
    )

    with open(output_dir / "subfolder" / "doc-1.ann") as f:
        exported_ann_text = f.read()

    assert_doc_write(exported_ann_text)



from typing import List

from pytest import fixture, mark

from edsnlp.pipelines.qualifiers.reported_speech import ReportedSpeech
from edsnlp.utils.examples import parse_example

examples: List[str] = [
    # (
    #     'Pas de critique de sa TS de nov 2020 "je '
    #     "<ent reported_speech_=REPORTED>regrette</ent> d'avoir raté\"."
    # ),
    # "Décrit un scénario d'<ent reported_speech_=REPORTED>IMV</ent>",
    (
        "Elles sont décrites par X.x. comme des appels à l'aide "
        "« La <ent reported_speech_=REPORTED>pendaison</ent> "
        "a permis mon hospitalisation »."
    ),
    (
        "Rapporte une tristesse de l'humeur avec des idées "
        "<ent reported_speech_=REPORTED>suicidiares</ent> à "
        "type de pendaison,"
    ),
    (
        "Décrit un fléchissement thymique depuis environ "
        "1 semaine avec idées suicidaires scénarisées "
        "(<ent reported_speech_=REPORTED>intoxication "
        "médicamenteuse volontaire)</ent>"
    ),
    (
        "Dit ne pas savoir comment elle est tombé. "
        'Minimise la chute. Dit que "ça arrive. Badaboum". '
        "Dit ne pas avoir fait <ent reported_speech_=REPORTED>IMV</ent>."
    ),
    (
        "Le patient parle \"d'en finir\", et dit qu'il a pensé "
        "plusieurs fois à se pendre où à se faire une "
        "<ent reported_speech_=REPORTED>phlébotomie</ent> "
        "lorsqu'il était dans la rue, diminution de ces "
        "idées noires depuis qu'il vit chez son fils"
    ),
]

# A long test to check leakage from one entity to the next.
examples.append(
    "le patient est admis pour coronavirus. il dit qu'il n'est "
    "pas <ent reported_speech=True>malade</ent>.\n"
    "les tests sont positifs.\n"
    "il est <ent reported_speech=False>malade</ent>"
)


@fixture
def reported_speech_factory(blank_nlp):

    default_config = dict(
        pseudo=None,
        preceding=None,
        following=None,
        quotation=None,
        verbs=None,
        attr="NORM",
        within_ents=False,
        explain=True,
    )

    def factory(on_ents_only, **kwargs):

        config = dict(**default_config)
        config.update(kwargs)

        return ReportedSpeech(
            nlp=blank_nlp,
            on_ents_only=on_ents_only,
            **config,
        )

    return factory


@mark.parametrize("on_ents_only", [True, False])
def test_reported_speech(blank_nlp, reported_speech_factory, on_ents_only):

    reported_speech = reported_speech_factory(on_ents_only=on_ents_only)

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = reported_speech(doc)

        for entity, ent in zip(entities, doc.ents):

            for modifier in entity.modifiers:

                assert getattr(ent._, modifier.key) == modifier.value

                if not on_ents_only:
                    for token in ent:
                        assert getattr(token._, modifier.key) == modifier.value


def test_reported_speech_within_ents(blank_nlp, reported_speech_factory):

    reported_speech = reported_speech_factory(on_ents_only=True, within_ents=True)

    examples = [
        "Le patient a une <ent reported_speech=True>"
        "fracture au tibias, il dit qu'il a mal</ent>."
    ]

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = reported_speech(doc)

        for entity, ent in zip(entities, doc.ents):

            for modifier in entity.modifiers:

                assert (
                    getattr(ent._, modifier.key) == modifier.value
                ), f"{modifier.key} labels don't match."



from edsnlp import registry

from .span_qualifier import TrainableSpanQualifier

create_component = registry.factory.register(
    "eds.span_qualifier",
    assigns=[],
    deprecated=[],
)(TrainableSpanQualifier)



from pytest import fixture

from edsnlp.matchers.utils import get_text
from edsnlp.pipelines.core.normalizer.accents.patterns import accents
from edsnlp.pipelines.core.normalizer.pollution.patterns import pollution
from edsnlp.pipelines.core.normalizer.quotes.patterns import quotes_and_apostrophes


@fixture
def text():
    return "L'aïeul ʺnˊest pas malade”, écrit-il. Fièvre NBNbWbWbNbWbNB jaune."


@fixture
def doc(nlp, text):
    return nlp(text)


def test_full_normalization(doc):
    norm = get_text(doc, attr="NORM", ignore_excluded=True)
    assert doc[1].norm_ == "aieul"
    assert norm == "l'aieul \"n'est pas malade\", ecrit-il. fievre jaune."


@fixture
def nlp_factory(blank_nlp):
    def f(a=False, lc=False, q=False, p=False):

        if a:
            a = dict(accents=accents)
        if q:
            q = dict(quotes=quotes_and_apostrophes)
        if p:
            p = dict(pollution=pollution)

        blank_nlp.add_pipe(
            "normalizer",
            config=dict(
                accents=a,
                lowercase=lc,
                quotes=q,
                pollution=p,
            ),
        )
        return blank_nlp

    return f


def test_normalization_accents(nlp_factory, text):

    nlp = nlp_factory(a=True)
    doc = nlp(text)

    norm = get_text(doc, attr="NORM", ignore_excluded=True)

    assert norm == "L'aieul ʺnˊest pas malade”, ecrit-il. Fievre NBNbWbWbNbWbNB jaune."


def test_normalization_spaces(nlp_factory, text):

    nlp = nlp_factory(a=True)
    doc = nlp("Phrase    avec des espaces \n et un retour à la ligne")

    tags = [t.tag_ for t in doc]
    assert tags == ["", "SPACE", "", "", "", "SPACE", "", "", "", "", "", ""]


def test_normalization_quotes(nlp_factory, text):

    nlp = nlp_factory(q=True)
    doc = nlp(text)

    norm = get_text(doc, attr="NORM", ignore_excluded=True)

    assert (
        norm == "L'aïeul \"n'est pas malade\", écrit-il. Fièvre NBNbWbWbNbWbNB jaune."
    )


def test_normalization_lowercase(nlp_factory, text):

    nlp = nlp_factory(lc=True)
    doc = nlp(text)

    norm = get_text(doc, attr="NORM", ignore_excluded=True)

    assert norm.startswith("l'aïeul")


def test_normalization_pollution(nlp_factory, text):

    nlp = nlp_factory(p=True)
    doc = nlp(text)

    norm = get_text(doc, attr="NORM", ignore_excluded=True)

    assert norm == "L'aïeul ʺnˊest pas malade”, écrit-il. Fièvre jaune."

    text2 = "Le jour de \n"
    text2 += "2/2Pat : <NOM> <Prenom> le <date> IPP <ipp> Intitulé RCP"
    text2 += " : Urologie HMN le <date>\nRéunion de Concertation"
    text2 += " Pluridisciplinaire\nHôpital Henri Mondor"
    text2_expected = (
        "Le jour de \nRéunion de Concertation Pluridisciplinaire\nHôpital Henri Mondor"
    )

    text3 = "Le jour de \n"
    text3 += "3/5CRH service ABC HC SOINS INTENSIFS CARDIOLOGIE - CARDIOLOGIE-2EME"
    text3 += " ETAGE-B    Pat.: Prenom NOM | M | 13/10/1789 | 8012345678 | xxxxxxxx \n"
    text3 += "consultation"
    text3_expected = "Le jour de \nconsultation"

    examples = [(text2, text2_expected), (text3, text3_expected)]

    for example, expected in examples:
        doc = nlp(example)
        norm = get_text(doc, attr="NORM", ignore_excluded=True)
        assert norm == expected



from libcpp cimport bool
from libcpp.set cimport set
from spacy.tokens.doc cimport Doc
from spacy.typedefs cimport attr_t


cdef class SentenceSegmenter(object):
    cdef str name
cdef class FastSentenceSegmenter(object):
    cdef bool ignore_excluded
    cdef attr_t newline_hash
    cdef attr_t excluded_hash
    cdef attr_t endline_hash
    cdef set[attr_t] punct_chars_hash
    cdef set[attr_t] capitalized_shapes_hash

    cdef void process(self, Doc doc) nogil



from typing import Dict, List, Union

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.core.matcher.matcher import GenericMatcher

from ...base import SpanSetterArg
from .patterns import patterns

DEFAULT_CONFIG = dict(
    attr="LOWER",
    ignore_excluded=False,
    ignore_space_tokens=False,
    patterns=patterns,
    label="covid",
    span_setter={"ents": True, "covid": True},
)


@registry.factory.register(
    "eds.covid",
    assigns=["doc.ents", "doc.spans"],
)
def create_component(
    nlp: PipelineProtocol,
    name: str = "covid",
    *,
    attr: Union[str, Dict[str, str]] = "LOWER",
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
    patterns: List[str] = patterns,
    label: str = "covid",
    span_setter: SpanSetterArg = {"ents": True, "covid": True},
):
    """
    The `eds.covid` pipeline component detects mentions of COVID19.

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.covid())

    text = "Le patient est admis pour une infection au coronavirus."

    doc = nlp(text)

    doc.ents
    # Out: (infection au coronavirus,)
    ```

    Parameters
    ----------
    nlp : PipelineProtocol
        spaCy `Language` object.
    name : str
        The name of the pipe
    attr : Union[str, Dict[str, str]]
        Attribute to match on, eg `TEXT`, `NORM`, etc.
    ignore_excluded : bool
        Whether to skip excluded tokens during matching.
    ignore_space_tokens : bool
        Whether to skip space tokens during matching.
    patterns : List[str]
        The regex pattern to use
    label : str
        Label to use for matches
    span_setter : SpanSetterArg
        How to set matches on the doc

    Returns
    -------
    GenericMatcher

    Authors and citation
    --------------------
    The `eds.covid` pipeline was developed by AP-HP's Data Science team.
    """

    return GenericMatcher(
        nlp=nlp,
        name=name,
        terms=None,
        regex={label: patterns},
        attr=attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
        span_setter=span_setter,
    )
